"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Aug 22, 2018

"""
from itertools import cycle
from typing import Callable
import warnings

import matplotlib.colors
import matplotlib.pyplot as plt
import matplotlib.ticker
import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy.stats import gaussian_kde
from sklearn import metrics

# TODO: propensity distribution using CDF (and not reflecting if so)
# TODO: consider making plots to not rely on pandas input (and can work more generally with numpy)?
# TODO: consider refactoring each type (family?) of plots to its own module (unify through __init__?)
# TODO: consider making plot module be class-based instead, taking its argument during init
#       and having a `plot()` interface

CONTINUOUS_ACCURACY_PLOT = "continuous_accuracy"
RESIDUALS_PLOT = "residuals"
COMMON_SUPPORT_PLOT = "common_support"
ROC_CURVE_PLOT = "roc_curve"
PR_CURVE_PLOT = "pr_curve"
CALIBRATION_PLOT = "calibration"
WEIGHT_DISTRIBUTION_PLOT = "weight_distribution"
COVARIATE_BALANCE_LOVE_PLOT = "covariate_balance_love"
COVARIATE_BALANCE_SLOPE_PLOT = "covariate_balance_slope"
COVARIATE_BALANCE_GENERIC_PLOT = "covariate_balance"

WeightPlotNames = frozenset({
    WEIGHT_DISTRIBUTION_PLOT,
    COVARIATE_BALANCE_LOVE_PLOT,
    COVARIATE_BALANCE_SLOPE_PLOT,
})

ContinuousOutputPlotNames = frozenset({
    CONTINUOUS_ACCURACY_PLOT,
    RESIDUALS_PLOT,
    COMMON_SUPPORT_PLOT,
})

BinaryOutputPlotNames = frozenset({CALIBRATION_PLOT, ROC_CURVE_PLOT, PR_CURVE_PLOT})


PropensityPlotNames = frozenset(BinaryOutputPlotNames | WeightPlotNames)



def lookup_name(name: str) -> Callable:
    """Lookup function for plot name.

    Canonical plot names are defined in this file as globals.
    Incorrect names will raise KeyError.

    Args:
        name (str): plot name to lookup

    Returns:
        Callable: plot function
    """
 
    return {
        CONTINUOUS_ACCURACY_PLOT: plot_continuous_prediction_accuracy_folds,
        RESIDUALS_PLOT: plot_residual_folds,
        COMMON_SUPPORT_PLOT: plot_counterfactual_common_support_folds,
        ROC_CURVE_PLOT: plot_roc_curve_folds,
        PR_CURVE_PLOT: plot_precision_recall_curve_folds,
        CALIBRATION_PLOT: plot_calibration_folds,
        WEIGHT_DISTRIBUTION_PLOT: plot_propensity_score_distribution_folds,
        COVARIATE_BALANCE_LOVE_PLOT: plot_mean_features_imbalance_love_folds,
        COVARIATE_BALANCE_SLOPE_PLOT: plot_mean_features_imbalance_slope_folds,
    }[name]


def _calculate_mutual_bins(x, y, bins="auto"):
    """
    A common support for two vectors.

    Args:
        x (pd.Series):
        y (pd.Series):
        bins: compatible with numpy's bins parameter.

    Returns:
        np.array: bins cutoffs.
    """
    data = np.append(x, y)
    bins = np.histogram(data, bins=bins)[1]
    return bins


def plot_counterfactual_common_support(prediction, a, ax=None):
    cv = [np.arange(a.shape[0])]
    ax = plot_counterfactual_common_support_folds([prediction], hue_by=a, cv=cv, ax=ax)
    return ax


def plot_counterfactual_common_support_folds(
    predictions, hue_by, cv, alpha_by_density=True, ax=None
):
    """Plot the scatter plot of y0 vs. y1 for multiple scoring results, colored by the treatment

    Args:
        predictions (list[pd.Series]): List, the size of number of folds, of outcome prediction values.
        hue_by (pd.Series): Group assignment (as in treatment assignment) of the entire dataset.
                            (indices from `cv` will be used to slice this vector)
        cv (list[np.array]): List, the size of number of folds, of row indices (as in iloc locations) - the indices
                             of samples participating the fold.
        alpha_by_density (bool): Whether to calculate points alpha value (transparent-opaque) with density estimation.
                                 This can take some time to compute for large number of points.
                                 If False, alpha calculation will be a simple fast heuristic.
        ax (plt.Axes): The axes on which the plot will be displayed. Optional.

    """
    effect_folds = [
        (prediction.iloc[:, 1] - prediction.iloc[:, 0]).mean()
        for prediction in predictions
    ]
    predictions = pd.concat(predictions)  # type: pd.DataFrame
    treatment = pd.concat([hue_by.iloc[fold_idx] for fold_idx in cv])  # type: pd.Series

    ax = _scatter_hue(
        predictions.iloc[:, 0],
        predictions.iloc[:, 1],
        treatment,
        alpha_by_density,
        ax=ax,
    )

    effect_label = rf"mean effect={np.mean(effect_folds):.2g}"
    effect_label += rf"$\pm${np.std(effect_folds):.2g}" if len(effect_folds) > 1 else ""
    ax.plot(
        [], [], color=ax.get_facecolor(), label=effect_label  # Use background color
    )
    _add_diagonal(ax)
    ax.legend(loc="best")
    ax.set_xlabel(r"Predicted $Y^0$")
    ax.set_ylabel(r"Predicted $Y^1$")
    ax.set_title("Predicted Common Support")
    return ax


def plot_continuous_prediction_accuracy(
    predictions, y, a, alpha_by_density=True, ax=None
):
    cv = [np.arange(a.shape[0])]
    ax = plot_continuous_prediction_accuracy_folds(
        [predictions], y, a, cv, alpha_by_density, ax=ax, plot_residuals=False
    )
    return ax


def plot_continuous_prediction_accuracy_folds(
    predictions, y, a, cv, alpha_by_density=True, plot_residuals=False, ax=None
):
    # Concatenate data across folds:
    treatments = []
    outcomes = []
    predictions_on_actual = []
    r2_scores = []
    for fold_prediction, fold_idx in zip(predictions, cv):
        fold_a = a.iloc[fold_idx]
        fold_y = y.iloc[fold_idx]
        if plot_residuals:
            fold_y = fold_y - fold_prediction

        r2_scores.append(metrics.r2_score(fold_y, fold_prediction))
        treatments.append(fold_a)
        outcomes.append(fold_y)
        predictions_on_actual.append(fold_prediction)

    treatments = pd.concat(treatments)  # type: pd.Series
    outcomes = pd.concat(outcomes)  # type: pd.Series
    predictions_on_actual = pd.concat(predictions_on_actual)  # type: pd.Series

    ax = _scatter_hue(predictions_on_actual, outcomes, treatments, alpha_by_density, ax)

    # R-squared label:
    if not plot_residuals:
        r2_label = rf"$R^2={np.mean(r2_scores):.2f}"
        r2_label += rf"\pm{np.std(r2_scores):.2f}$" if len(r2_scores) > 1 else "$"
        ax.plot(
            [], [], color=ax.get_facecolor(), label=r2_label
        )  # invisible color so as to not show line in legend
        _add_diagonal(ax)

    ax.legend(loc="best")
    ax.set_xlabel("Predicted values")
    ax.set_ylabel("Prediction residuals" if plot_residuals else "True values")
    ax.set_title("Residual Plot" if plot_residuals else "Continuous Accuracy Plot")
    return ax


def plot_residual_folds(predictions, y, a, cv, alpha_by_density=True, ax=None):
    ax = plot_continuous_prediction_accuracy_folds(
        predictions, y, a, cv, alpha_by_density, plot_residuals=True, ax=ax
    )
    ax.axhline(0.0, linestyle="--", color="grey", zorder=0, alpha=0.75)
    return ax


def plot_residual(predictions, y, a, alpha_by_density=True, ax=None):
    cv = [np.arange(a.shape[0])]
    ax = plot_residual_folds([predictions], y, a, cv, alpha_by_density, ax)
    return ax


def _scatter_hue(x, y, hue, alpha_by_density=True, ax=None):
    ax = ax or plt.gca()

    points_rgba = (
        _get_alpha_per_point_with_density(X=[x, y], hue=hue)
        if alpha_by_density
        else None
    )

    for i, treatment_val in enumerate(np.sort(np.unique(hue))):
        idx_mask = hue == treatment_val  # type: pd.Series
        cur_color = (
            points_rgba.loc[idx_mask].values if points_rgba is not None else None
        )
        cur_alpha = np.clip(10 / np.sqrt(idx_mask.sum()), 0.01, 1)

        ax.scatter(
            x=x.loc[idx_mask],
            y=y.loc[idx_mask],
            alpha=cur_alpha if points_rgba is None else None,
            facecolor=cur_color,
            edgecolors="none",
            label=f"treatment={treatment_val}",
        )
    return ax


def _get_alpha_per_point_with_density(X, hue, min_alpha_bound=0.3, max_alpha_bound=1.0):
    """
    Matplotlib does not support pointwise alpha values (rather, constant value for an entire plt.plot()).
    This function will utilize a supported pointwise color-scheme, using rgba, and passing the individual alpha values
    as the 4th dimension ('a') of the rgba.

    Args:
        X: in a form compatible with statsmodels' KDEMultivariate (list of pd.Series, or pd.DataFrame)
        hue (pd.Series): A vector with group assignment for each point in x.
        min_alpha_bound (float | None): Value between 0 and 1, used to linearly rescale the alpha values.
                                        If None, rescale is avoided.
                                        Default of 0.3, since lower values are usually too unobservable.
        max_alpha_bound (float | None): Value between 0 and 1, used to linearly rescale the alpha values.
                                        If None, rescale is avoided.

    Returns:

    """
    points_rgba = pd.DataFrame(index=hue.index, columns=list("rgba"), dtype=np.float64)

    # Calculate alpha for each point based on its density:
    kde = sm.nonparametric.KDEMultivariate(data=X, var_type="cc", bw="normal_reference")
    # kde.bw = kde.bw * 0.5         # Rescale bandwidth to be narrower
    points_density = kde.pdf(X)
    # Invert values - the denser the point -> the lower its alpha (more transparent)
    points_alpha = 1 / points_density
    if (min_alpha_bound is not None) and (max_alpha_bound is not None):
        #   Rescale alphas (linearly) to the range of 0.3 to 1:
        points_alpha = min_alpha_bound + (max_alpha_bound - min_alpha_bound) * (
            (points_alpha - points_alpha.min())
            / (points_alpha.max() - points_alpha.min())
        )
    points_rgba["a"] = points_alpha  # Assign the alpha values

    for i, hue_val in enumerate(np.sort(np.unique(hue))):
        idx_mask = hue == hue_val
        cur_color = f"C{i}"  # Cycle through the colors
        cur_color = matplotlib.colors.to_rgb(
            cur_color
        )  # Get RGB value of the current color
        points_rgba.loc[
            idx_mask, ["r", "g", "b"]
        ] = cur_color  # Assign that constant RGB val for all current points

    return points_rgba


def plot_calibration_folds(
    predictions,
    targets,
    cv,
    n_bins=10,
    plot_se=True,
    plot_rug=False,
    plot_histogram=False,
    quantile=False,
    ax=None,
):
    """Plot calibration curves for multiple models (presumably in folds)

    Args:
        predictions (list[pd.Series]): list (each entry of a fold) of arrays - probability ("scores") predictions.
        targets (pd.Series): true labels to calibrate against on the overall data (not divided to folds).
        cv (list[np.array]):
        n_bins (int): number of bins to evaluate in the plot
        plot_se (bool): Whether to plot standard errors around the mean bin-probability estimation.
        plot_rug:
        plot_histogram:
        quantile (bool): If true, the binning of the calibration curve is by quantiles. Default is false
        ax (plt.Axes): Optional

    Note:
        One of plot_propensity or plot_model must be True.

    Returns:

    """
    for i, idx_fold in enumerate(cv):
        predictions_fold = predictions[i]
        target_fold = targets.iloc[idx_fold]

        ax = _plot_calibration_single(
            y_true=target_fold,
            y_prob=predictions_fold,
            n_bins=n_bins,
            plot_diagonal=False,
            plot_se=plot_se,
            plot_rug=plot_rug,
            plot_histogram=plot_histogram,
            quantile=quantile,
            label=f"fold {i}",
            ax=ax,
        )
    _add_diagonal(ax)
    ax.legend(loc="best")
    # ax.set_title("{} Calibration".format("Propensity" if y is None else "Outcome"))
    ax.set_title("Calibration")
    return ax


def plot_calibration(
    predictions,
    targets,
    n_bins=10,
    plot_se=True,
    plot_rug=False,
    plot_histogram=True,
    quantile=False,
    ax=None,
):
    cv = [np.arange(predictions.shape[0])]
    return plot_calibration_folds(
        [predictions],
        targets,
        cv=cv,
        n_bins=n_bins,
        plot_se=plot_se,
        plot_rug=plot_rug,
        plot_histogram=plot_histogram,
        quantile=quantile,
        ax=ax,
    )


def _plot_calibration_single(
    y_true,
    y_prob,
    n_bins=10,
    plot_diagonal=True,
    plot_se=True,
    plot_rug=False,
    plot_histogram=False,
    quantile=False,
    label=None,
    ax=None,
):
    """Plot a calibration curve showing how well y_prob predicts the probability of a binary outcome y

    The standard deviation of a binomial distribution p(1-p)/sqrt(n) is used to calculate the values for which p
    would be one standard deviation away. This means we are looking for
    r +/- sqrt(r(1-r)/n) = p
    This provides a cubic equation for r whose solution is
    r = (2np+1 +/- sqrt(4np(1-p)+1)) / (2n+2)

    Args:
        y_prob (pd.Series):
        y_true (pd.Series):
        n_bins (int): the number of bins to use for the calibration plot
        plot_se (bool): Whether to plot standard errors around the
                        mean bin-probability estimation.
        plot_diagonal (bool): Whether to plot a diagonal line or not.
        plot_rug (bool): Whether to plot rug of the prediction
        plot_histogram (bool): Whether to plot histogram at the background.
        quantile (bool): If False specifies equal sized bins,
                         if True splits the probabilities into n_bins quantiles.
        ax (plt.Axes):
        label(str): The label for the plotted line

    Returns:

    """
    ax = ax or plt.gca()
    if quantile:
        bins = np.unique(
            np.percentile(y_prob, np.linspace(0, 100, n_bins + 1).astype(int))
        )
        # in case all values of y_prob are the same
        bins = bins if len(bins) > 1 else np.concatenate([bins, bins])
        bins[-1] += 1e-8
        prob_true, prob_pred, counts = calibration_curve(y_true, y_prob, bins=bins)
    else:
        prob_true, prob_pred, counts = calibration_curve(y_true, y_prob, bins=n_bins)
        bins = np.linspace(0.0, 1.0 + 1e-8, n_bins + 1)

    if plot_rug:
        ax.plot(y_prob, np.full_like(y_prob, 0.01), "|", color="black", alpha=0.7)

    line_color = None
    if plot_histogram:
        hist_line = ax.plot(
            bins, (counts / counts.sum()), drawstyle="steps-post", alpha=0.8
        )
        hist_line = hist_line[0]
        # keep histogram behind any new lines that are plotted after it.
        hist_line.set_zorder(2)
        # if plotting hist, keep track of color to use in the line to be plotted
        line_color = hist_line.get_color()

    if plot_diagonal:
        _add_diagonal(ax)
    lines = ax.plot(prob_pred, prob_true, "s-", color=line_color, label=label)

    # Plot standard error:
    if plot_se:
        disc = (4 * counts * prob_true) * (1 - prob_true) + 1
        upper = (2 * counts * prob_true + 1 + np.sqrt(disc)) / (2 * counts + 2)
        lower = (2 * counts * prob_true + 1 - np.sqrt(disc)) / (2 * counts + 2)
        ax.fill_between(
            x=prob_pred, y1=lower, y2=upper, color=lines[-1].get_color(), alpha=0.5
        )

    ax.set_xlabel("Predicted probability")
    ax.set_ylabel("Observed probability")
    return ax


def calibration_curve(y_true, y_prob, bins=5):
    """
    Compute calibration curve of a classifier given its scores output and true label assignment.

    Args:
        y_true (pd.Series): True binary label assignment.
        y_prob (pd.Series): Predicted probability of each sample being the positive label.
        bins (int | list | np.ndarray | pd.Series):
            If int, it defines the number of equal-width bins in the
            given range (5, by default).
            If bins a sequence, it defines the bin edges, including the
            rightmost edge, allowing for non-uniform bin widths.

    Returns:
        (pd.Series, pd.Series, pd.Series): empirical_prob, predicted_prob, bin_counts
            empirical_prob: The fraction of positive labels in each bins
            predicted_prob: The average of predicted probability in each bin
            bin_counts: The number of samples fallen in each bin

    References:
        [1] Zadrozny, B., & Elkan, C. (2002, July).
            Transforming classifier scores into accurate multiclass probability estimates

    """
    # Get binning out of provided bins
    if type(bins) is int:
        bins = np.linspace(0.0, 1.0 + 1e-8, bins + 1)
    elif hasattr(bins, "__len__") and not isinstance(bins, str):  # Some sort of vector
        bins = np.sort(np.ravel(bins))
        if y_prob.max() > bins.max() or y_prob.min() < bins.min():
            raise ValueError("y_prob has values outside the provided bins")
    else:
        raise TypeError("bins must either be an integer or a sequence of scalars")

    bin_of_samples = pd.cut(y_prob, bins, labels=np.arange(len(bins) - 1)).astype(int)
    predicted_prob = y_prob.groupby(bin_of_samples).mean()
    empirical_prob = y_true.groupby(bin_of_samples).mean()
    bin_counts = bin_of_samples.value_counts(sort=False)
    return empirical_prob, predicted_prob, bin_counts


def plot_roc_curve_folds(
    curve_data, ax=None, plot_folds=False, label_folds=False, label_std=False, **kwargs
):
    num_of_curves = len(curve_data.keys())
    color_list = [f"C{_}" for _ in range(num_of_curves)]

    for (curve_name, curve_data), color in zip(curve_data.items(), color_list):
        fprs = curve_data["FPR"]
        tprs = curve_data["TPR"]
        aucs = curve_data["AUC"]

        ax = _plot_single_performance_curve(
            fprs,
            tprs,
            aucs,
            "AUC",
            color,
            curve_name,
            label_std,
            label_folds,
            plot_folds,
            num_of_curves != 1,
            ax,
        )
    # Plot chance curve:
    ax.plot(
        [0, 1], [0, 1], linestyle="--", lw=2, color="black", label="Chance", alpha=0.8
    )

    ax.set_xlim(left=-0.05, right=1.05)
    ax.set_ylim(bottom=-0.05, top=1.05)
    ax.set_xlabel("False Positive Rate")
    ax.set_ylabel("True Positive Rate")
    ax.set_title("ROC Curve")
    ax.legend(loc="lower right")
    return ax


def plot_precision_recall_curve_folds(
    curve_data, ax=None, plot_folds=False, label_folds=False, label_std=False, **kwargs
):
    # TODO: Check why it does not end at class prevalence (for recall=1.0)
    num_of_curves = len(curve_data.keys())
    color_list = [f"C{_}" for _ in range(num_of_curves)]

    pos_class_prevalence = curve_data.pop("prevalence", None)

    for (curve_name, curve_data), color in zip(curve_data.items(), color_list):
        recalls = curve_data["Recall"]
        precisions = curve_data["Precision"]
        aps = curve_data["AP"]

        ax = _plot_single_performance_curve(
            recalls,
            precisions,
            aps,
            "AP",
            color,
            curve_name,
            label_std,
            label_folds,
            plot_folds,
            num_of_curves != 1,
            ax,
        )
    # Plot chance curve:
    if pos_class_prevalence is not None:
        ax.plot(
            [0, 1],
            [pos_class_prevalence, pos_class_prevalence],
            linestyle="--",
            lw=2,
            color="black",
            label="Chance",
            alpha=0.8,
        )

    ax.set_xlim(left=-0.05, right=1.05)
    ax.set_ylim(bottom=-0.05, top=1.05)
    ax.set_xlabel("Recall")
    ax.set_ylabel("Precision")
    ax.set_title("PR Curve")
    ax.legend(loc="lower left")
    return ax


def _plot_single_performance_curve(
    xs,
    ys,
    areas,
    areas_type,
    color="C0",
    curve_name="",
    label_std=False,
    label_folds=False,
    plot_folds=False,
    colored_folds=False,
    ax=None,
):
    ax = ax or plt.gca()
    assert len(xs) == len(ys) == len(areas)

    n_folds = len(xs)
    x_domain = np.linspace(0, 1, 100)
    ys_interp = []
    for i in range(n_folds):
        if areas_type == "AP":  # precision/recall need to be reversed for interpolation
            ys_interp.append(np.interp(x_domain, xs[i][::-1], ys[i][::-1]))
        else:
            ys_interp.append(np.interp(x_domain, xs[i], ys[i]))
            ys_interp[-1][0] = 0.0
        area = areas[i]

        folds_label = f"Fold {i} ({areas_type} = {area:.2f})" if label_folds else None
        if plot_folds:
            # use multiple colors if plotting only one stratum
            folds_color = None if colored_folds else color
            ax.plot(xs[i], ys[i], lw=1, alpha=0.3, color=folds_color, label=folds_label)

    # Plot main (folds average) curve
    mean_ys = np.nanmean(ys_interp, axis=0)
    # if areas_type == "AUC":
    #     mean_ys[-1] = 1.0
    mean_area = np.nanmean(areas)
    std_area = np.nanstd(areas)
    ax.plot(
        x_domain,
        mean_ys,
        color=color,
        label=rf"{curve_name} ({areas_type} = {mean_area:.2f} $\pm$ {std_area:.2f})",
        lw=2,
        alpha=0.9,
    )

    # Plot uncertainty around main curve:
    ys_std = np.std(ys_interp, axis=0)
    upper_ys = np.minimum(mean_ys + ys_std, 1)
    lower_ys = np.maximum(mean_ys - ys_std, 0)
    std_label = r"$\pm$ 1 std. dev." if label_std else None
    ax.fill_between(
        x_domain, lower_ys, upper_ys, color=color, alpha=0.2, label=std_label
    )

    return ax


def plot_propensity_score_distribution(
    propensity,
    treatment,
    reflect=True,
    kde=False,
    cumulative=False,
    norm_hist=True,
    ax=None,
):
    """
    Plot the distribution of propensity score

    Args:
        propensity (pd.Series):
        treatment (pd.Series):
        reflect (bool): Whether to plot second treatment group on the opposite sides of the x-axis.
                        This can only work if there are exactly two groups.
        kde (bool): Whether to plot kernel density estimation
        cumulative (bool): Whether to plot cumulative distribution.
        norm_hist (bool): If False - use raw counts on the y-axis.
                          If kde=True, then norm_hist should be True as well.
        ax (plt.Axes | None):

    Returns:

    """
    # assert propensity.index.symmetric_difference(a.index).size == 0
    ax = ax or plt.gca()
    if kde and not norm_hist:
        warnings.warn(
            "kde=True and norm_hist=False is not supported. Forcing norm_hist from False to True."
        )
        norm_hist = True
    bins = np.histogram(propensity, bins="auto")[1]
    plot_params = dict(bins=bins, density=norm_hist, alpha=0.5, cumulative=cumulative)

    unique_treatments = np.sort(np.unique(treatment))
    for treatment_number, treatment_value in enumerate(unique_treatments):
        cur_propensity = propensity.loc[treatment == treatment_value]
        cur_color = f"C{treatment_number}"
        ax.hist(
            cur_propensity,
            label=f"treatment = {treatment_value}",
            color=[cur_color],
            **plot_params,
        )
        if kde:
            cur_kde = gaussian_kde(cur_propensity)
            min_support = max(0, cur_propensity.values.min() - cur_kde.factor)
            max_support = min(1, cur_propensity.values.max() + cur_kde.factor)
            X_plot = np.linspace(min_support, max_support, 200)
            if cumulative:
                density = np.array(
                    [cur_kde.integrate_box_1d(X_plot[0], x_i) for x_i in X_plot]
                )
                ax.plot(
                    X_plot,
                    density,
                    color=cur_color,
                )
            else:
                ax.plot(
                    X_plot,
                    cur_kde.pdf(X_plot),
                    color=cur_color,
                )
    if reflect:
        if len(unique_treatments) != 2:
            raise ValueError(
                "Reflecting density across X axis can only be done for two groups. "
                "This one has {}".format(len(unique_treatments))
            )
        # Update line:
        if kde:
            last_line = ax.get_lines()[-1]
            last_line.set_ydata(-1 * last_line.get_ydata())
        # Update histogram bars:
        idx_of_first_hist_rect = [patch.get_label() for patch in ax.patches].index(
            f"treatment = {unique_treatments[-1]}"
        )
        for patch in ax.patches[idx_of_first_hist_rect:]:
            patch.set_height(-1 * patch.get_height())

        # Re-set the view of axes:
        ax.relim()
        ax.autoscale()
        # Remove negation sign from lower y-axis:
        ax.yaxis.set_major_formatter(
            matplotlib.ticker.FuncFormatter(
                lambda x, pos: str(x) if x >= 0 else str(-x)
            )
        )

    ax.legend(loc="best")
    x_type = (
        "Propensity" if propensity.between(0, 1, inclusive="both").all() else "Weights"
    )
    ax.set_xlabel(x_type)
    y_type = "Probability density" if norm_hist else "Counts"
    ax.set_ylabel(y_type)
    ax.set_title(f"{x_type} Distribution")
    return ax


def plot_propensity_score_distribution_folds(
    predictions,
    hue_by,
    cv,
    reflect=True,
    kde=False,
    cumulative=False,
    norm_hist=True,
    ax=None,
):
    """

    Args:
        predictions (list[pd.Series]):
        X (pd.DataFrame):
        hue_by (pd.Series):
        y (pd.Series):
        cv (list[np.array]):
        reflect (bool): Whether to plot second treatment group on the opposite sides of the x-axis.
                        This can only work if there are exactly two groups.
        kde (bool): Whether to plot kernel density estimation
        cumulative (bool): Whether to plot cumulative distribution.
        norm_hist (bool): If False - use raw counts on the y-axis.
                          If kde=True, then norm_hist should be True as well.
        ax (plt.Axis):

    Returns:

    """
    propensity = pd.concat(predictions)  # type: pd.Series
    # treatment = hue_by         # if train phase then there will  be no duplication of records.
    treatment = pd.concat([hue_by.iloc[fold_idx] for fold_idx in cv])  # type: pd.Series
    ax = plot_propensity_score_distribution(
        propensity,
        treatment,
        reflect=reflect,
        kde=kde,
        cumulative=cumulative,
        norm_hist=norm_hist,
        ax=ax,
    )
    return ax


def plot_mean_features_imbalance_love_folds(
    table1_folds,
    cv=None,
    aggregate_folds=True,
    thresh=None,
    plot_semi_grid=True,
    ax=None,
):
    method_pretty_name = {
        "smd": "Standard Mean Difference",
        "abs_smd": "Absolute Standard Mean Difference",
        "ks": "Kolmogorov-Smirnov",
    }
    ax = ax or plt.gca()

    # Aggregate across folds. This will be used to determine order, and extreme values.
    # Use this groupby trick: https://stackoverflow.com/a/25058102
    aggregated_table1 = pd.concat(table1_folds)  # type: pd.DataFrame
    aggregated_table1 = aggregated_table1.groupby(aggregated_table1.index)

    order = aggregated_table1.mean().sort_values(by="unweighted", ascending=True).index

    if aggregate_folds:
        # place in iterable to make compatible with input
        table1_folds = [aggregated_table1.mean()]

    # Plot:
    for table1 in table1_folds:
        color_cycle = cycle(plt.rcParams["axes.prop_cycle"].by_key()["color"])
        marker_cycle = cycle(["o", "^", "P", "s", "*"])
        for col_name, col_data in table1.items():
            col_data = col_data.loc[order]  # unified order
            ax.scatter(
                col_data,
                order,
                label=col_name,
                marker=next(marker_cycle),
                color=next(color_cycle),
                zorder=0,
            )

    # Plot line connecting the dots: (before plotting dots so they would be underneath the dots)
    if plot_semi_grid:
        if aggregate_folds:
            h_max = aggregated_table1.mean().max(axis="columns")
            h_min = aggregated_table1.mean().min(axis="columns")
        else:
            h_max = aggregated_table1.max().max(axis="columns")
            h_min = aggregated_table1.min().min(axis="columns")
        ax.hlines(
            h_min.index,
            xmin=h_min,
            xmax=h_max,  # ax.hlines(order, xmin=h_min, xmax=h_max,
            colors="grey",
            linestyles="dashed",
            zorder=1,
            label=None,
        )

    # Plot vertical threshold line
    if thresh is not None:
        ax.axvline(thresh, color="grey", linestyle="--", zorder=2)
        if aggregated_table1.min().min().min() < 0:
            # There are negative values, plot the minus of threshold and adjust x-limits to be symmetric:
            ax.axvline(-thresh, color="grey", linestyle="--", zorder=2)
            ax.set_xlim(-np.max(np.abs(ax.get_xlim())), np.max(np.abs(ax.get_xlim())))

    # # If too many features, remove their tick labels:
    fig = ax.get_figure()
    ax_pixel_height = (
        ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted()).height
        * fig.dpi
    )
    # 10 is hypothesized to be font size + 3 pt. margin
    if ax_pixel_height / order.size < 10 + 3:
        ax.set_yticklabels([])  # Too many y-ticks for axis size, remove them.

    ax.set_xlabel(
        method_pretty_name.get(
            table1_folds[0].columns.name, table1_folds[0].columns.name
        )
    )
    ax.set_ylabel("Covariates")
    ax.legend(loc="lower right")
    return ax


def plot_mean_features_imbalance_scatter_plot(
    table1_folds,
    aggregate_folds=True,
    thresh=None,
    ax=None,
):
    # get current axes
    ax = ax or plt.gca()

    method_pretty_name = {
        "smd": "Standard Mean Difference",
        "abs_smd": "Absolute Standard Mean Difference",
        "ks": "Kolmogorov-Smirnov",
    }
    # Aggregate across folds. This will be used to determine order, and extreme values.
    # Use this groupby trick: https://stackoverflow.com/a/25058102
    aggregated_table1 = pd.concat(table1_folds)  # type: pd.DataFrame
    aggregated_table1 = aggregated_table1.groupby(aggregated_table1.index)

    if aggregate_folds:
        table1_folds = [aggregated_table1.mean()]
    
    # Plot:

    for table1 in table1_folds:
        
        # setting different marker shapes for each fold in aggregated_foldes == False
        marker_cycle = cycle(["o", "^", "P", "s", "*"]) 

        # find index of features that are above threshold 
        violating = table1["weighted"] > thresh
        # determain color for dot on plot 
        color = violating.replace({False: "C0", True: "C1"})
        
        
        ax.scatter( 
            x=table1['unweighted'],
            y=table1['weighted'],
            marker=next(marker_cycle),
            color=color
        )
        for covariate_name, covariate_diff in table1.loc[violating].iterrows():
            ax.text(
                x=covariate_diff["unweighted"],
                y=covariate_diff["weighted"],
                s=covariate_name,
                horizontalalignment="left",
            )
            
    # Plot vertical and horizontal threshold line
    if thresh is not None:
        ax.axvline(thresh, color="grey", linestyle="--", zorder=2)
        ax.axhline(thresh, color="grey", linestyle="--", zorder=2)
        # There are negative values, plot the minus of threshold   
        if aggregated_table1.min().min().min() < 0:
            # There are negative values, plot the minus of threshold and adjust x-limits to be symmetric:
            ax.axvline(-thresh, color="grey", linestyle="--", zorder=2)
            ax.axhline(-thresh, color="grey", linestyle="--", zorder=2)
            ax.set_xlim(-np.max(np.abs(ax.get_xlim())), np.max(np.abs(ax.get_xlim())))

    
    # adding labels 

    metric_name = table1_folds[0].columns.name
    metric_name = method_pretty_name.get(metric_name, metric_name)

    ax.set_xlabel(f'Unweighted [{metric_name}]')
    ax.set_ylabel(f'Weighted [{metric_name}]')

    return ax 


def plot_mean_features_imbalance_slope_folds(
    table1_folds, cv=None, thresh=None, ax=None
):
    method_pretty_name = {
        "smd": "Standard Mean Difference",
        "abs_smd": "Absolute Standard Mean Difference",
        "ks": "Kolmogorov-Smirnov",
    }
    # ax = ax or plt.gca()

    # Aggregate across folds. This will be used to determine order, and extreme values.
    # Use this groupby trick: https://stackoverflow.com/a/25058102
    aggregated_table1 = pd.concat(table1_folds)  # type: pd.DataFrame
    aggregated_table1 = aggregated_table1.groupby(aggregated_table1.index)
    aggregated_table1 = aggregated_table1.mean()

    # Reorder:
    aggregated_table1 = aggregated_table1.sort_values(by="unweighted", ascending=True)

    # Slope graph:
    ax = slope_graph(
        left=aggregated_table1["unweighted"],
        right=aggregated_table1["weighted"],
        thresh=thresh,
        ax=ax,
    )

    ax.set_ylabel(
        method_pretty_name.get(
            table1_folds[0].columns.name, table1_folds[0].columns.name
        )
    )
    # ax.legend(loc="upper right")
    return ax


def slope_graph(
    left, right, thresh=None, color_below="C0", color_above="C1", marker="o", ax=None
):
    ax = ax or plt.gca()
    left_xtick = left.name or "unweighted"
    right_xtick = right.name or "weighted"

    if thresh is not None:
        ax.axhline(thresh, color="grey", linestyle="--", zorder=2)
        # There are negative values, plot the minus of threshold
        if left.min() < 0 or right.min() < 0:
            ax.axhline(-thresh, color="grey", linestyle="--", zorder=2)
    else:
        thresh = np.nan  # will be now used to compare against values

    for idx in left.index:
        cur_left = left[idx]
        cur_right = right[idx]
        # make default color_below if thresh is nan
        cur_color = color_above if cur_right > thresh else color_below

        ax.plot(
            [left_xtick, right_xtick],
            [cur_left, cur_right],
            label=None,
            color=cur_color,
            marker=marker,
        )
        if cur_right > thresh:
            ax.text(x=1.01, y=cur_right, s=idx, horizontalalignment="left")

    # Place y-tick labels on both sides:
    ax.tick_params(left=True, labelleft=True, right=True, labelright=True)
    return ax


def get_subplots(n_features, max_cols=5, fig_size=(16, 16), sharex=False, sharey=False):
    """Initializes the grid of subplots and returns the axes

    Args:
        n_features (int): The total number of features to plot
        max_cols (int): The maximal number of figures in each row of figures
        fig_size (tuple[int, int]): Passed on to matplotlib
        sharex (str|bool): will be passed to subplots
        sharey (str|bool): will be passed to subplots

    Returns:
        tuple[Figure, np.ndarray]: the figure and the array of axes

    """
    # try to make the plots as square as possible
    ncols = min(int(np.round(np.sqrt(n_features))), max_cols)
    nrows = int(np.ceil(n_features / ncols))
    fig, axes = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        figsize=fig_size,
        sharex=sharex,
        sharey=sharey,
        squeeze=False,
    )
    return fig, axes


def _add_diagonal(
    ax, fraction=0.04, label="x=y", color="grey", linestyle="--", zorder=1
):
    diagonal = [
        max(ax.get_xlim()[0], ax.get_ylim()[0]),
        min(ax.get_xlim()[1], ax.get_ylim()[1]),
    ]
    # ax.plot(diagonal, diagonal, color="black", label="x=y")  # plot diagonal
    lim_range_frac = (
        np.array([np.diff(ax.get_xlim()), np.diff(ax.get_ylim())]) * fraction
    )
    while np.any(np.abs(np.diff(diagonal)[0]) < lim_range_frac):
        if np.abs(np.diff(diagonal)[0]) < lim_range_frac[0]:
            ax.set_ylim(
                *(
                    ax.get_ylim()
                    + np.diff(ax.get_ylim()) * [-fraction / 2, fraction / 2]
                )
            )
        if np.abs(np.diff(diagonal)[0]) < lim_range_frac[1]:
            ax.set_xlim(
                *(
                    ax.get_xlim()
                    + np.diff(ax.get_xlim()) * [-fraction / 2, fraction / 2]
                )
            )
        diagonal = [
            max(ax.get_xlim()[0], ax.get_ylim()[0]),
            min(ax.get_xlim()[1], ax.get_ylim()[1]),
        ]
        # ax.plot(diagonal, diagonal, color="black")  # extend diagonal
        lim_range_frac = (
            np.array([np.diff(ax.get_xlim()), np.diff(ax.get_ylim())]) * fraction
        )

    # plot diagonal
    ax.plot(
        diagonal, diagonal, color=color, label=label, linestyle=linestyle, zorder=zorder
    )


WEIGHT_PLOTS = {
    "weight_distribution": plot_propensity_score_distribution_folds,
    "covariate_balance_love": plot_mean_features_imbalance_love_folds,
    "covariate_balance_slope": plot_mean_features_imbalance_slope_folds,
}

OUTCOME_PLOTS = {
    "continuous_accuracy": plot_continuous_prediction_accuracy_folds,
    "residuals": plot_residual_folds,
    "common_support": plot_counterfactual_common_support_folds,
}

SHARED_PLOTS = {
    "roc_curve": plot_roc_curve_folds,
    "pr_curve": plot_precision_recall_curve_folds,
    "calibration": plot_calibration_folds,
}
OUTCOME_PLOTS.update(SHARED_PLOTS)
WEIGHT_PLOTS.update(SHARED_PLOTS)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on April 4, 2021
"""
import warnings
import pandas as pd
import numpy as np
from sklearn.utils.multiclass import type_of_target

from .base_estimator import IndividualOutcomeEstimator

from ..utils.crossfit import cross_fitting
from ..utils import general_tools as g_tools


class VotingEstimator:
    def __init__(self, estimators):
        self.estimators = estimators

    def _stack_predictions(self, X):
        """Collect results from clf.predict_proba or ref.predict calls."""
        if hasattr(self.estimators[0], "predict_proba"):
            predictions = [est.predict_proba(X)[:, 1] for est in self.estimators]
        else:
            predictions = [est.predict(X) for est in self.estimators]
        predictions = np.stack(predictions, axis=-1)
        return predictions

    def predict(self, X):
        """Aggregate results of different estimators"""
        predicted_target = self._stack_predictions(X)
        averaged_target = np.average(predicted_target, axis=-1)
        return pd.Series(averaged_target, index=X.index)


class RLearner(IndividualOutcomeEstimator):
    """
    Given the measured outcome Y, the assignment A, and the coefficients X
    calculate an R-learner estimator of the effect of the treatment
    Let e(X) be the estimated propensity score and m(X) is the estimated outcome
    (E[Y|X]) by an estimator, then the R-learner minimize the following:
        ||Y - m(X) - (A-e(X))\tau(X)||^2_2 + lambda (\tau)
    where \tau(X) is a conditional average treatment effect and
    lambda is a regularize coefficient.

    If the effect_model is Linear, than minimizing squared loss with
    the target variable (Y-m(X)) and the features (A-e(X))X,
    otherwise it corresponds to a weighted regression problem,
    where the weights are (A-e(X))**2. This can be used with any scikit-learn
    regressor that accepts sample weights

    References:
    Nie, X., & Wager, S.(2017).
    Quasi - oracle estimation of heterogeneous treatment effects
    https://arxiv.org/abs/1712.04912

    Chernozhukov, V., et al. (2018).
    Double/debiased machine learning for treatment and structural parameters.‏
    https://academic.oup.com/ectj/article/21/1/C1/5056401
    """

    def __init__(
        self,
        effect_model,
        outcome_model,
        treatment_model,
        outcome_covariates=None,
        treatment_covariates=None,
        effect_covariates=None,
        n_splits=5,
        refit=True,
        caliper=1e-6,
        non_parametric=False,
    ):
        """
        Args:
            effect_model: An sklearn model that estimate that estimate
                the conditional average treatment effect \tau(X)
            outcome_model: An sklearn model that estimate the
                regressor Y|X (without the treatment).
                Note: it is recommended to use a regressor, even for binary outcome.
            treatment_model: An sklearn model that estimate the treatment model
                or the probability to be treated, i.e A|X or P(A=1|X)
            outcome_covariates (array): Covariates to use for the outcome model.
                If None - all covariates passed will be used.
                Either list of column names or boolean mask.
            treatment_covariates (array): Covariates to use for treatment model.
                If None - all covariates passed will be used.
                Either list of column names or boolean mask.
            effect_covariates (array): Covariates to use for the effect model.
                If None - all covariates passed will be used.
                Either list of column names or boolean mask.
            n_splits (int): number of sample-splitting in the cross-fitting procedure
            refit (bool): if True - Nuisance models are fitted over the whole
                training set, otherwise Nuisance models are fitted per folds
            non_parametric(bool): if True - the effect_model is estimated as
                weighted regression task, otherwise the effect_model is
                considered linear.
        """

        super(RLearner, self).__init__(lambda **x: None)  # Dummy initialization
        delattr(self, "learner")  # To remove the learner attribute a
        # IndividualOutcomeEstimator has
        self.effect_model = effect_model
        self.outcome_model = outcome_model
        self.treatment_model = treatment_model
        self.outcome_covariates = outcome_covariates
        self.treatment_covariates = treatment_covariates
        self.effect_covariates = effect_covariates
        self.n_splits = n_splits
        self.refit = refit
        self.caliper = caliper
        self.non_parametric = non_parametric

    def __repr__(self):
        repr_string = (
            "{cls_name}(\n "
            "outcome_model={outcome},\n "
            "treatment_model={treatment},\n "
            "effect_model={effect})"
        ).format(
            cls_name=self.__class__.__name__,
            outcome=self.outcome_model,
            treatment=self.treatment_model,
            effect=self.effect_model,
        )
        return repr_string

    def _extract_outcome_model_data(self, X):
        outcome_covariates = self.outcome_covariates or X.columns
        X_outcome = X[outcome_covariates]
        return X_outcome

    def _extract_treatment_model_data(self, X):
        treatment_covariates = self.treatment_covariates or X.columns
        X_treatment = X[treatment_covariates]
        return X_treatment

    def _extract_effect_model_data(self, X):
        X_effect = pd.Series(data=1, index=X.index, name="intercept_").to_frame()
        if not isinstance(self.effect_covariates, list) or self.effect_covariates:
            effect_covariates = self.effect_covariates or X.columns
            X_effect = pd.concat([X_effect, X[effect_covariates]], axis="columns")
        return X_effect

    def _prepare_data(self, X):
        """
        Extract the relevant parts for outcome model, treatment model and
        effect model for the entire data matrix

        Args:
            X (pd.DataFrame): Covariate matrix of size
                              (num_subjects, num_features).
        Returns:
            (pd.DataFrame, pd.DataFrame, pd.DataFrame): X_outcome, X_treatment, X_effect
                Data matrix for outcome model, treatment model and data effect model
        """
        X_outcome = self._extract_outcome_model_data(X)
        X_treatment = self._extract_treatment_model_data(X)
        X_effect = self._extract_effect_model_data(X)
        return X_outcome, X_treatment, X_effect

    def _fit_and_predict_model(self, model, X, target, predict_proba):
        """
        fit outcome model (E[Y|X]) or treatment model (E[A|X] or P[A=1|X])
        and produce predictions on held-out data subsets
        Args:
            model (object): Treatment or outcome model
            X (pd.DataFrame): Covariate matrix of size
                              (num_subjects, num_features).
            target (pd.Series): Observed target (outcome or treatment)
                                of size (num_subjects,).
            predict_proba (bool): if True predict probabilities with predict_proba,
                                  otherwise use predict
        Returns:
            pd.Series: cross-fitted prediction of the outcome
        """
        target_pred, estimators = cross_fitting(
            model, X, target, n_splits=self.n_splits, predict_proba=predict_proba
        )
        target_pred = pd.Series(data=target_pred, index=X.index)
        if self.refit:
            estimators = [model.fit(X, target)]
        return target_pred, estimators

    def _fit_and_predict_nuisance(self, X_outcome, X_treatment, a, y):
        """
        fit the nuisance models and return residuals
        Args:
            X_outcome (pd.DataFrame): Covariate matrix of size
                (num_subjects, num_features).
            X_treatment (pd.DataFrame): Covariate matrix of size
                (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).
        Returns:
            pd.Series, pd.Series:
                residuals of the outcome model, residuals of the treatment model
        """
        # residuals of the outcome model
        pred_y, outcome_model_ = self._fit_and_predict_model(
            self.outcome_model, X_outcome, y, predict_proba=False
        )
        res_y = y - pred_y
        self.outcome_model_ = VotingEstimator(outcome_model_)

        # residuals of the treatment model
        pred_a, treatment_model_ = self._fit_and_predict_model(
            self.treatment_model, X_treatment, a, predict_proba=self.binary_treatment_
        )
        res_a = a - pred_a
        self.treatment_model_ = VotingEstimator(treatment_model_)

        return res_y, res_a

    def estimate_individual_effect(self, X):
        """
        Predict the individual treatment effect
        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).

        Returns:
            pd.Series: The series is a vector in size (num_subjects) that
                contains the estimated treatment effect, each row is an individual
        """
        X_effect = self._extract_effect_model_data(X)
        return pd.Series(self.effect_model.predict(X_effect), index=X.index)

    def estimate_individual_outcome(self, X, a, treatment_values=None,
                                    predict_proba=False):
        """
        Estimating corrected individual counterfactual outcomes.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            treatment_values (Any): Desired treatment value/s to use when
                estimating the counterfactual outcome.
                If not supplied, calculates for all available treatment values.
            predict_proba: IGNORED.
                Not used, present for API consistency by convention.
        Returns:
            pd.DataFrame: DataFrame which columns are treatment values and rows
                are individuals: each column is a vector size (num_samples,)
                that contains the estimated outcome for each individual under
                the treatment value in the corresponding key.
        """
        if (not self.binary_treatment_) & (treatment_values is None):
            raise ValueError(
                "The individual_outcome cannot be represented "
                "for a continuous treatment effect.\n Choose "
                'treatment in "treatment_values" '
            )
        X_outcome = self._extract_outcome_model_data(X)
        X_treatment = self._extract_treatment_model_data(X)

        outcome_pred = self.outcome_model_.predict(X_outcome)
        treatment_pred = self.treatment_model_.predict(X_treatment)
        effect_pred = self.estimate_individual_effect(X)

        treatment_values = g_tools.get_iterable_treatment_values(treatment_values, a)
        individual_cf = {
            treatment_value: (treatment_value - treatment_pred) * effect_pred
            + outcome_pred
            for treatment_value in treatment_values
        }
        return pd.DataFrame(individual_cf)

    def _fit_linear_effect_model(self, X, res_a, res_y):
        """
        fit a linear effect model without fitting an intercept
        Args:
            X (pd.DataFrame): Covariate matrix of size
                              (num_subjects, num_features).
            res_a (pd.Series): residuals of the treatment model
            res_y (pd.Series): residuals of the outcome model
        """
        if hasattr(self.effect_model, "fit_intercept"):
            if self.effect_model.fit_intercept:
                self.effect_model.fit_intercept = False
                warnings.warn(
                    "The effect model forces intercept estimation as an "
                    "additional coefficient. Therefore, the explicit "
                    "`fit_intercept` attribute was set to False."
                )
        else:
            warnings.warn(
                "`non_parametric=False` was passed and `effect_model` "
                "does not have a `fit_intercept` attribute and therefore "
                "does not seem to be scikit-learn LinearModel. "
                "Ignore this warning if you are sure `effect_model` "
                "is a linear model."
            )
        # multiply each row by a scalar
        X_tilde = X * np.array(res_a)[:, np.newaxis]
        self.effect_model.fit(X_tilde, res_y)

    def _fit_non_parametric_effect_model(self, X, res_a, res_y, caliper):
        """
        fit a non-parametric effect model as a weighted regression.
        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            res_a (pd.Series): residuals of the treatment model
            res_y (pd.Series): residuals of the outcome model
            caliper (None | float): minimal value of treatment-probability residual
        """
        if hasattr(self.effect_model, "fit_intercept"):
            warnings.warn(
                "`non_parametric=True` was passed, but `effect_model` has a "
                "`fit_intercept` attributes and therefore seems to be a"
                "scikit-learn LinearModel. Ignore this warning if you're sure"
                "`effect_model` is a nonparametric model"
            )

        caliper_ = caliper or self.caliper
        if caliper_ is not None:
            cond_clipping = np.abs(res_a) < caliper_
            res_a.loc[cond_clipping] = caliper_ * np.sign(res_a.loc[cond_clipping])
        self.effect_model.fit(X, (res_y / res_a), sample_weight=res_a ** 2)

    def fit(self, X, a, y, caliper=None):
        """
        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).
            caliper (None | float): minimal value of treatment-probability residual.
                used to avoid division by zero when fitting the effect-model.
                If None - no clipping is done.
                The caliper is irrelevant if the effect_model is Linear.
        """
        self.binary_treatment_ = type_of_target(a) == "binary"
        X_outcome, X_treatment, X_effect = self._prepare_data(X)
        res_y, res_a = self._fit_and_predict_nuisance(X_outcome, X_treatment, a, y)
        if self.non_parametric:
            self._fit_non_parametric_effect_model(X_effect, res_a, res_y, caliper)
        else:
            self._fit_linear_effect_model(X_effect, res_a, res_y)
        return self



"""Causal Survival Analysis Models"""

from .univariate_curve_fitter import UnivariateCurveFitter
from .regression_curve_fitter import RegressionCurveFitter
from .marginal_survival import MarginalSurvival
from .weighted_survival import WeightedSurvival
from .standardized_survival import StandardizedSurvival
from .weighted_standardized_survival import WeightedStandardizedSurvival

__all__ = [
    "UnivariateCurveFitter",
    "RegressionCurveFitter",
    "MarginalSurvival",
    "WeightedSurvival",
    "StandardizedSurvival",
    "WeightedStandardizedSurvival",
]



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Jul 03, 2017

"""
from __future__ import division

import unittest
from builtins import range, zip
from collections import namedtuple

import numpy as np
import pandas as pd

import causallib.simulation.CausalSimulator3 as CS3m  # CS3 module    to test helper functions
from causallib.simulation.CausalSimulator3 import CausalSimulator3 as CS3


class CS3TestCase(unittest.TestCase):
    def setUp(self):
        # DEFAULTS:
        Params = namedtuple("Params", ["topology", "var_types", "outcome_types", "link_types", "prob_cat", "snr",
                                       "treatment_importance", "effect_sizes"])

        topology = np.zeros((4, 4), dtype=np.bool_)  # topology[i,j] if node j is a parent of node i
        topology[1, 0] = topology[2, 0] = topology[2, 1] = topology[3, 1] = topology[3, 2] = True
        self.no_X = Params(topology=topology, var_types=["hidden", "covariate", "treatment", "outcome"],
                           outcome_types=["continuous"], link_types=['linear'] * 4,
                           prob_cat=[None, None, [0.5, 0.5], None],
                           snr=0.5, treatment_importance=0.8, effect_sizes=None)
        self.NUM_SAMPLES = 2000
        self.X_NUM_SAMPLES = 2500

        self.X_GIVEN = pd.DataFrame(np.random.RandomState(0).normal(size=(self.X_NUM_SAMPLES, 5)),
                                    columns=["x{}".format(i) for i in range(5)])
        topology = np.zeros((5 + 4, 5 + 4), dtype=np.bool_)  # topology[i,j] if node j is a parent of node i
        topology[1, 0] = topology[2, 0] = topology[2, 1] = topology[3, 1] = topology[3, 2] = True
        topology[1, 4] = topology[2, 4] = topology[2, 6] = topology[3, 6] = topology[3, 7] = True
        var_names = pd.Index(list(range(4))).append(self.X_GIVEN.columns)
        self.with_X = Params(topology=topology,
                             var_types=pd.Series(data=["hidden", "covariate", "treatment", "outcome"] +
                                                      ["hidden"] + ["covariate"] * 4, index=var_names),
                             outcome_types=["continuous"], link_types=pd.Series(data=['linear'] * 9, index=var_names),
                             prob_cat=pd.Series([None, None, [0.5, 0.5], None] + [None] * 5, index=var_names),
                             snr=0.5, treatment_importance=0.8, effect_sizes=None)

    def dependency_from_topology(self, sim):
        """
        Tests to see that the matrix topology is well converted into graph dependencies.
        """
        for i, row in enumerate(sim.topology):
            i = sim.var_names[i]
            parents_mat_idx = set(sim.var_names[np.where(row)[0]])
            parents_grp_idx = set(sim.graph_topology.predecessors(i))
            self.assertEqual(parents_mat_idx, parents_grp_idx,
                             msg="variable {i} had these predecessors in matrix topology: {mat} "
                                 "and these predecessors in graph topology: {grp}".format(i=i, mat=parents_mat_idx,
                                                                                          grp=parents_grp_idx))

    def test_bad_input(self):
        # lengths:
        # with self.assertRaises(ValueError) as assert_checker:
        var_types = ["covariate", "treatment", "outcome"]
        self.assertRaises(ValueError, CS3, topology=self.no_X.topology, var_types=var_types,
                          prob_categories=self.no_X.prob_cat, link_types=self.no_X.link_types,
                          treatment_importances=self.no_X.treatment_importance, snr=self.no_X.snr,
                          outcome_types=self.no_X.outcome_types, effect_sizes=self.no_X.effect_sizes)

        # outcome has more than one treatment predecessor:
        var_types = ["covariate", "treatment", "treatment", "outcome"]
        self.assertRaises(ValueError, CS3, topology=self.no_X.topology, var_types=var_types,
                          prob_categories=self.no_X.prob_cat, link_types=self.no_X.link_types,
                          treatment_importances=self.no_X.treatment_importance, snr=self.no_X.snr,
                          outcome_types=self.no_X.outcome_types, effect_sizes=self.no_X.effect_sizes)

        # No valid link type:
        self.assertRaises(ValueError, CS3, topology=self.no_X.topology, var_types=self.no_X.var_types,
                          prob_categories=self.no_X.prob_cat,
                          link_types=["linear", "linear", "linear", "leniar"],
                          treatment_importances=self.no_X.treatment_importance, snr=self.no_X.snr,
                          outcome_types=self.no_X.outcome_types, effect_sizes=self.no_X.effect_sizes)

        # No valid treatment method:
        self.assertRaises(ValueError, CS3, topology=self.no_X.topology, var_types=self.no_X.var_types,
                          prob_categories=self.no_X.prob_cat, link_types=self.no_X.link_types,
                          treatment_importances=self.no_X.treatment_importance, snr=self.no_X.snr,
                          outcome_types=self.no_X.outcome_types, effect_sizes=self.no_X.effect_sizes,
                          treatment_methods="rndom")

        # lengths:
        self.assertRaises(ValueError, CS3, topology=self.no_X.topology, var_types=self.no_X.var_types,
                          prob_categories=self.no_X.prob_cat, link_types=self.no_X.link_types,
                          treatment_importances=self.no_X.treatment_importance, snr=[0, 1],
                          outcome_types=self.no_X.outcome_types, effect_sizes=self.no_X.effect_sizes,
                          treatment_methods="gaussian")
        self.assertRaises(ValueError, CS3, topology=self.no_X.topology, var_types=self.no_X.var_types,
                          prob_categories=self.no_X.prob_cat, link_types=self.no_X.link_types,
                          treatment_importances=[0.5, 0.5],
                          snr=self.no_X.snr, treatment_methods="gaussian",
                          outcome_types=self.no_X.outcome_types, effect_sizes=self.no_X.effect_sizes)

        # no generation input:
        sim = CS3(topology=self.no_X.topology, var_types=self.no_X.var_types, prob_categories=self.no_X.prob_cat,
                  link_types=self.no_X.link_types, treatment_importances=self.no_X.treatment_importance,
                  snr=self.no_X.snr, outcome_types=self.no_X.outcome_types, effect_sizes=self.no_X.effect_sizes)
        self.assertRaises(ValueError, sim.generate_data)

        # categorical treatment:
        self.assertRaises(ValueError, sim.generate_treatment_col,
                          X_parents=pd.DataFrame([None]), link_type=None, snr=1, prob_category=None)

        # wrong probabilities:
        prob_cat = [[0.5, -0.5, 1], None, [0.5, 0.5], None]
        sim = CS3(topology=self.no_X.topology, var_types=self.no_X.var_types, prob_categories=prob_cat,
                  link_types=self.no_X.link_types, treatment_importances=self.no_X.treatment_importance,
                  snr=self.no_X.snr, outcome_types=self.no_X.outcome_types, effect_sizes=self.no_X.effect_sizes)
        self.assertRaises(ValueError, sim.generate_data, num_samples=100)
        prob_cat = [None, None, [0.5, 0.6], None]
        sim = CS3(topology=self.no_X.topology, var_types=self.no_X.var_types, prob_categories=prob_cat,
                  link_types=self.no_X.link_types, treatment_importances=self.no_X.treatment_importance,
                  snr=self.no_X.snr, outcome_types=self.no_X.outcome_types, effect_sizes=self.no_X.effect_sizes)
        self.assertRaises(ValueError, sim.generate_data, num_samples=100)

    def test_different_types_of_paramaters(self):
        """
        Tests to see what happens when supplying parameters of different types (lists, dicts, arrays, etc.)
        """
        topology = np.zeros((5, 5), dtype=np.bool_)  # topology[i,j] if node j is a parent of node i
        topology[1, 0] = topology[2, 0] = topology[2, 1] = topology[3, 1] = topology[3, 2] = topology[3, 4] = True
        var_types = ["hidden", "covariate", "treatment", "outcome", "covariate"]
        sim = CS3(topology=topology, var_types=var_types, prob_categories=self.no_X.prob_cat + [None],
                  link_types=None, treatment_importances=pd.Series(data=0.7, index=[2]),
                  outcome_types={3: "continuous"}, snr=0.5,
                  # effect_sizes={2: 0.8},
                  effect_sizes=[0.8],
                  treatment_methods=["gaussian"])
        self.assertTrue(all(sim.link_types == "linear"))
        self.assertEqual(len(sim.link_types), 5)
        self.assertTrue(all([x == 0.7 for x in sim.treatment_importances]))
        self.assertTrue(sim.outcome_types.equals(pd.Series({3: "continuous"})))
        self.assertTrue(all(sim.snr == 0.5))
        self.assertEqual(len(sim.snr), 5)
        self.assertTrue(sim.effect_sizes.equals(pd.Series({3: 0.8})))

    # ### Test to see that structure make sense ### #
    def test_dependency_from_topology(self):
        """
        Tests to see that the matrix topology is well converted into graph dependencies for with and without dataset.
        """
        sim = CS3(topology=self.no_X.topology, var_types=self.no_X.var_types, prob_categories=self.no_X.prob_cat,
                  link_types=self.no_X.link_types, treatment_importances=self.no_X.treatment_importance,
                  outcome_types=self.no_X.outcome_types, snr=self.no_X.snr, effect_sizes=self.no_X.effect_sizes)
        self.dependency_from_topology(sim)
        sim = CS3(topology=self.with_X.topology, var_types=self.with_X.var_types,
                  prob_categories=self.with_X.prob_cat, link_types=self.with_X.link_types,
                  treatment_importances=self.with_X.treatment_importance, outcome_types=self.with_X.outcome_types,
                  snr=self.with_X.snr, effect_sizes=self.with_X.effect_sizes)
        self.dependency_from_topology(sim)

    def test_dataset_size(self):
        """
        Tests to see the the size of the generated dataset is ok under several configurations
        """
        # No given X, all non-hidden
        var_types = ["covariate", "covariate", "treatment", "outcome"]
        sim = CS3(topology=self.no_X.topology, var_types=var_types, prob_categories=self.no_X.prob_cat,
                  link_types=self.no_X.link_types, treatment_importances=self.no_X.treatment_importance,
                  outcome_types=self.no_X.outcome_types, snr=self.no_X.snr, effect_sizes=self.no_X.effect_sizes)
        X, prop, cf = sim.generate_data(num_samples=self.NUM_SAMPLES)
        self.assertEqual(X.shape, (self.NUM_SAMPLES, 4),
                         msg="Generated dataset shape is {X} "
                             "but supposed to be {supp}".format(X=X.shape, supp=(self.NUM_SAMPLES, 4)))
        self.assertEqual(prop.shape, (self.NUM_SAMPLES, 2),
                         msg="Generated propensity shape is {X} "
                             "but supposed to be {supp}".format(X=prop.shape, supp=(self.NUM_SAMPLES, 2)))
        self.assertEqual(cf.shape, (self.NUM_SAMPLES, 2),
                         msg="number of generated counterfactuals is {X} "
                             "but supposed to be {supp}".format(X=cf.shape, supp=(self.NUM_SAMPLES, 2)))
        df_obs, df_cf = sim.format_for_training(X, prop, cf)
        self.assertEqual(df_obs.shape, (self.NUM_SAMPLES, 4),
                         msg="Generated dataset shape is {X} "
                             "but supposed to be {supp}".format(X=df_obs.shape, supp=(self.NUM_SAMPLES, 4)))

        # No given X, with hidden
        var_types = ["hidden", "hidden", "treatment", "outcome"]
        sim = CS3(topology=self.no_X.topology, var_types=var_types, prob_categories=self.no_X.prob_cat,
                  link_types=self.no_X.link_types, treatment_importances=self.no_X.treatment_importance,
                  outcome_types=self.no_X.outcome_types, snr=self.no_X.snr, effect_sizes=self.no_X.effect_sizes)
        X, prop, cf = sim.generate_data(num_samples=self.NUM_SAMPLES)
        df_obs, df_cf = sim.format_for_training(X, prop, cf)
        self.assertEqual(df_obs.shape, (self.NUM_SAMPLES, 2),
                         msg="Generated dataset shape is {X} "
                             "but supposed to be {supp}".format(X=df_obs.shape, supp=(self.NUM_SAMPLES, 2)))

        # Given X, with hidden vars
        sim = CS3(topology=self.with_X.topology, var_types=self.with_X.var_types,
                  prob_categories=self.with_X.prob_cat, link_types=self.with_X.link_types,
                  treatment_importances=self.with_X.treatment_importance, outcome_types=self.with_X.outcome_types,
                  snr=self.with_X.snr, effect_sizes=self.with_X.effect_sizes)
        X, prop, cf = sim.generate_data(X_given=self.X_GIVEN)
        self.assertEqual(X.shape, (self.X_NUM_SAMPLES, 9),
                         msg="Generated dataset shape is {X} "
                             "but supposed to be {supp}".format(X=X.shape, supp=(self.X_NUM_SAMPLES, 9)))
        self.assertEqual(prop.shape, (self.X_NUM_SAMPLES, 2),
                         msg="Generated propensity shape is {X} "
                             "but supposed to be {supp}".format(X=prop.shape, supp=(self.X_NUM_SAMPLES, 2)))
        self.assertEqual(cf.shape, (self.X_NUM_SAMPLES, 2),
                         msg="Number of counterfactuals generated is {X} "
                             "but supposed to be {supp}".format(X=cf.shape, supp=(self.X_NUM_SAMPLES, 2)))
        df_obs, df_cf = sim.format_for_training(X, prop, cf)
        self.assertEqual(df_obs.shape, (self.X_NUM_SAMPLES, 7),
                         msg="Generated dataset shape is {X} "
                             "but supposed to be {supp}".format(X=df_obs.shape, supp=(self.X_NUM_SAMPLES, 7)))

    def test_multi_treatment_outcome(self):
        topology = np.zeros((6, 6), dtype=bool)
        topology[2, 0] = topology[3, 0] = topology[2, 1] = topology[3, 1] = topology[4, 2] = topology[5, 3] = True
        var_types = ["covariate", "covariate", "treatment", "treatment", "outcome", "outcome"]
        link_types = ["linear"] * 6
        prob_cat = [None] * 6
        prob_cat[2] = prob_cat[3] = [0.5, 0.5]
        sim = CS3(topology=topology, var_types=var_types, prob_categories=prob_cat, link_types=link_types,
                  snr=self.no_X.snr, treatment_importances=self.no_X.treatment_importance,
                  outcome_types=["continuous", "continuous"], effect_sizes=self.no_X.effect_sizes)
        X, prop, cf = sim.generate_data(num_samples=self.NUM_SAMPLES)
        self.assertEqual(prop.shape, (self.NUM_SAMPLES, 4),
                         msg="Generated propensity shape is {X} "
                             "but supposed to be {supp}".format(X=prop.shape, supp=(self.NUM_SAMPLES, 4)))
        self.assertEqual(cf.shape, (self.NUM_SAMPLES, 4),
                         msg="Number of generated counterfactuals is {X} "
                             "but supposed to be {supp}".format(X=cf.shape, supp=(self.NUM_SAMPLES, 4)))

    def test_multi_categorical_treatment(self):
        t_probs = pd.Series([0.2, 0.2, 0.1, 0.5])
        prob_cat = [None, None, t_probs, None]
        treatment_methods = ["quantile_gauss_fit", "odds_ratio"]
        decimals = [1, 1]
        for treatment_method, decimal in zip(treatment_methods, decimals):
            sim = CS3(topology=self.no_X.topology, var_types=self.no_X.var_types, prob_categories=prob_cat,
                      link_types=self.no_X.link_types, treatment_importances=self.no_X.treatment_importance,
                      outcome_types=self.no_X.outcome_types, snr=self.no_X.snr, effect_sizes=self.no_X.effect_sizes,
                      treatment_methods=treatment_method)
            n = self.NUM_SAMPLES * 50
            X, prop, cf = sim.generate_data(num_samples=n)
            np.testing.assert_array_almost_equal(prop.sum(axis="columns"), np.ones(n),
                                                 err_msg="multi-categorical preopensities of treatment method {method} "
                                                         "does not sum to 1".format(method=treatment_method))
            np.testing.assert_array_almost_equal(np.array(X[2].value_counts(normalize=True) - t_probs), np.zeros(4),
                                                 decimal=decimal,
                                                 err_msg="treatment method {method} does not produce proportions as "
                                                         "required".format(method=treatment_method))

    # ### Test helpers ### #
    def test_random_topology_generation(self):
        # ### without given variables: ### #
        T, var_types = CS3m.generate_random_topology(n_covariates=4, p=0.4, n_treatments=2, n_outcomes=2,
                                                     n_censoring=0, given_vars=[], p_hidden=0)
        # test output structure:
        self.assertEqual(T.shape[0], T.shape[1], msg="Graph has no square shape")
        self.assertEqual(T.shape[0], 8,
                         msg="Number of Graph variables {emp} "
                             "does not match it supposed number {sup}".format(emp=T.shape[0], sup=8))
        self.assertEqual(T.shape[0], var_types.size)
        # test number of variables of each type matches:
        self.assertEqual(sum(var_types == "covariate"), 4)
        self.assertEqual(sum(var_types == "treatment"), 2)
        self.assertEqual(sum(var_types == "outcome"), 2)
        self.assertEqual(sum(var_types == "hidden"), 0)
        self.assertEqual(sum(var_types == "censor"), 0)
        # test that each treatment is coupled with one outcome:
        self.assertEqual(all(T.loc[var_types == "outcome", var_types == "treatment"].sum(axis=1) == np.array([1, 1])),
                         True, msg="each outcome variable does not have exactly one predecessor treatment variable")

        # ### with hidden variables and censor variables: ### #
        T, var_types = CS3m.generate_random_topology(n_covariates=100, p=0.4, n_treatments=2, n_outcomes=2,
                                                     n_censoring=2, given_vars=[], p_hidden=0.4)
        # test output structure:
        self.assertEqual(T.shape[0], T.shape[1], msg="Graph has no square shape")
        self.assertEqual(T.shape[0], 106,
                         msg="Number of Graph variables {t} does not match it supposed number {s}".format(t=T.shape[0],
                                                                                                          s=106))
        self.assertEqual(T.shape[0], var_types.size)
        # test number of variables of each type matches:
        self.assertEqual(sum(var_types == "censor"), 2)
        hist = var_types.value_counts()
        self.assertAlmostEqual(hist["hidden"] / 100.0, 0.4, delta=1e-2)

        # graph = nx.from_numpy_matrix(T.values.transpose(), create_using=nx.DiGraph())
        # ### with given variables: ### #
        X = pd.DataFrame(np.random.RandomState(0).normal(size=(4800, 5)))
        T, var_types = CS3m.generate_random_topology(n_covariates=4, p=0.4, n_treatments=2, n_outcomes=2,
                                                     n_censoring=0, given_vars=X.columns, p_hidden=0)
        self.assertEqual(sum(var_types == "covariate"), 9)
        # test that given variable has no predecessors:
        np.testing.assert_array_equal(T.loc[X.columns, :].sum(axis="columns"), np.zeros(5))

        # Test for DAGness:
        from networkx import DiGraph, from_numpy_matrix, is_directed_acyclic_graph
        NUM_TESTS = 50
        for test in range(NUM_TESTS):
            n_cov = np.random.randint(low=10, high=100)
            p = np.random.rand()  # type: float
            n_tre_out = np.random.randint(low=1, high=4)
            n_cen = np.random.randint(low=0, high=n_tre_out)
            T, _ = CS3m.generate_random_topology(n_covariates=n_cov, p=p, n_treatments=n_tre_out, n_outcomes=n_tre_out,
                                                 n_censoring=n_cen, given_vars=[], p_hidden=0)
            G = from_numpy_matrix(T.values.transpose(), create_using=DiGraph())
            res = is_directed_acyclic_graph(G)
            self.assertTrue(res)

    # ### Test to see that values make sense ### #
    def test_categorical_proportions(self):
        probs = np.array([0.25, 0.25, 0.5])
        prob_cat = self.no_X.prob_cat
        prob_cat[1] = probs
        sim = CS3(topology=self.no_X.topology, var_types=self.no_X.var_types, prob_categories=prob_cat,
                  link_types=self.no_X.link_types, treatment_importances=self.no_X.treatment_importance,
                  outcome_types=self.no_X.outcome_types, snr=self.no_X.snr, effect_sizes=self.no_X.effect_sizes)
        X, prop, cf = sim.generate_data(num_samples=self.NUM_SAMPLES * 10, random_seed=0)
        # hist = np.array(X.loc[:, 1].value_counts(normalize=True))
        hist = X.loc[:, 1].value_counts(normalize=True)
        probs = pd.Series(probs)
        np.testing.assert_array_almost_equal(probs - hist, pd.Series(data=0, index=probs.index), decimal=2,
                                             err_msg="Empirical distribution {emp} of categories "
                                                     "is too far from desired distribution {des}".format(emp=hist,
                                                                                                         des=probs))

    def test_linear_linking(self):
        topology = np.zeros((3, 3), dtype=bool)
        topology[2, 0] = topology[2, 1] = True
        var_types = ["covariate", "treatment", "outcome"]
        snr = 1
        prob_cat = [None, [0.5, 0.5], None]
        treatment_importance = None
        sim = CS3(topology=topology, var_types=var_types, prob_categories=prob_cat,
                  link_types="linear", treatment_importances=treatment_importance,
                  outcome_types=self.no_X.outcome_types, snr=snr, effect_sizes=self.no_X.effect_sizes)
        X, prop, cf = sim.generate_data(num_samples=self.NUM_SAMPLES)

        singular_values = np.linalg.svd(X.values, compute_uv=False)
        eps = 1e-10
        rank = np.sum(singular_values > eps)
        self.assertEqual(rank, 2,
                         msg="discovered rank of matrix is {emp} instead of {des}."
                             "so the linear linking does not work properly".format(emp=rank, des=2))

    def test_treatment_logistic(self):
        topology = np.zeros((6, 6), dtype=bool)
        topology[2, 0] = topology[3, 0] = topology[2, 1] = topology[3, 1] = topology[4, 2] = topology[5, 3] = True
        var_types = ["covariate", "covariate", "treatment", "treatment", "outcome", "outcome"]
        link_types = ["linear"] * 6
        prob_cat = [None] * 6
        prob_cat[2] = prob_cat[3] = [0.5, 0.5]
        sim = CS3(topology=topology, var_types=var_types, prob_categories=prob_cat, link_types=link_types,
                  snr=self.no_X.snr, treatment_importances=self.no_X.treatment_importance,
                  outcome_types=["continuous", "continuous"], effect_sizes=self.no_X.effect_sizes,
                  treatment_methods=["logistic", "logistic"])
        X, prop, cf = sim.generate_data(num_samples=self.NUM_SAMPLES)
        # TODO: how to check logistic? (using logit?)

    def test_treatment_random(self):
        topology = np.zeros((6, 6), dtype=bool)
        topology[2, 0] = topology[3, 0] = topology[2, 1] = topology[3, 1] = topology[4, 2] = topology[5, 3] = True
        var_types = ["covariate", "covariate", "treatment", "treatment", "outcome", "outcome"]
        link_types = ["linear"] * 6
        prob_cat = [None] * 6
        prob_cat[2] = [0.5, 0.5]
        prob_cat[3] = [0.2, 0.8]
        sim = CS3(topology=topology, var_types=var_types, prob_categories=prob_cat, link_types=link_types,
                  snr=self.no_X.snr, treatment_importances=self.no_X.treatment_importance,
                  outcome_types=["continuous", "continuous"], effect_sizes=self.no_X.effect_sizes,
                  treatment_methods=["random", "random"])
        num_samples = self.NUM_SAMPLES * 10
        X, prop, cf = sim.generate_data(num_samples=num_samples)
        np.testing.assert_array_equal(prop[2][1], [0.5] * num_samples)
        np.testing.assert_array_equal(prop[3][1], [0.8] * num_samples)
        hist = X[2].value_counts(normalize=True)
        np.testing.assert_almost_equal(hist, [0.5, 0.5], decimal=2)
        hist = X[3].value_counts(normalize=True)
        np.testing.assert_almost_equal(hist.sort_index(), [0.2, 0.8], decimal=2)

    def test_treatment_gaussian(self):
        topology = np.zeros((6, 6), dtype=bool)
        topology[2, 0] = topology[3, 0] = topology[2, 1] = topology[3, 1] = topology[4, 2] = topology[5, 3] = True
        var_types = ["covariate", "covariate", "treatment", "treatment", "outcome", "outcome"]
        link_types = ["linear"] * 6
        prob_cat = [None] * 6
        prob_cat[2] = [0.5, 0.5]
        prob_cat[3] = [0.2, 0.8]
        sim = CS3(topology=topology, var_types=var_types, prob_categories=prob_cat, link_types=link_types,
                  snr=self.no_X.snr, treatment_importances=self.no_X.treatment_importance,
                  outcome_types=["continuous", "continuous"], effect_sizes=self.no_X.effect_sizes,
                  treatment_methods=["random", "random"])
        num_samples = self.NUM_SAMPLES * 5
        X, prop, cf = sim.generate_data(num_samples=num_samples)
        np.testing.assert_array_equal(prop[2][1], [0.5] * num_samples)
        np.testing.assert_array_equal(prop[3][1], [0.8] * num_samples)
        hist = X[2].value_counts(normalize=True)
        np.testing.assert_almost_equal(hist, [0.5, 0.5], decimal=2)
        hist = X[3].value_counts(normalize=True)
        np.testing.assert_almost_equal(hist.sort_index(), [0.2, 0.8], decimal=2)

        sim = CS3(topology=self.no_X.topology, var_types=self.no_X.var_types, prob_categories=self.no_X.prob_cat,
                  link_types=self.no_X.link_types, treatment_importances=self.no_X.treatment_importance,
                  outcome_types=self.no_X.outcome_types, snr=self.no_X.snr, effect_sizes=self.no_X.effect_sizes)
        X, prop, cf = sim.generate_data(num_samples=num_samples, random_seed=0)
        hist = X.loc[:, 2].value_counts(normalize=True)  # treatment
        np.testing.assert_almost_equal(hist, [0.5, 0.5], decimal=2)

    def test_survival_outcome(self):
        topology = np.zeros((5, 5), dtype=bool)
        topology[3, 0] = topology[4, 0] = topology[3, 1] = topology[3, 2] = topology[4, 2] = topology[4, 3] = True
        var_types = ["covariate", "covariate", "hidden", "treatment", "outcome"]
        link_types = ["linear"] * 5
        prob_cat = [None] * 5
        prob_cat[3] = [0.2, 0.8]
        outcome_type = "survival"
        snr = 0.95
        treatment_importance = None
        treatment_method = "logistic"
        survival_distribution = "expon"
        survival_baseline = 0.8
        sim = CS3(topology=topology, var_types=var_types, prob_categories=prob_cat, link_types=link_types,
                  snr=snr, treatment_importances=treatment_importance,
                  outcome_types=outcome_type, effect_sizes=self.no_X.effect_sizes,
                  treatment_methods=treatment_method,
                  survival_distribution=survival_distribution, survival_baseline=survival_baseline)
        num_samples = self.NUM_SAMPLES
        X, prop, cf = sim.generate_data(num_samples=num_samples)
        # TODO: how to test this?

    def test_censoring(self):
        # survival censor
        topology = np.zeros((5, 5), dtype=bool)
        topology[2, 0] = topology[3, 0] = topology[4, 0] = topology[2, 1] = topology[3, 1] = topology[4, 1] = topology[
            3, 2] = topology[4, 2] = topology[4, 3] = True  # make censor be dependent like the outcome
        var_types = ["covariate", "covariate", "treatment", "censor", "outcome"]
        link_types = ["linear"] * 5
        prob_cat = [None, None, [0.2, 0.8], [0.85, 0.15], None]
        outcome_type = "survival"
        snr = 0.95
        treatment_importance = None
        treatment_method = "logistic"
        survival_distribution = "expon"
        survival_baseline = 0.8
        sim = CS3(topology=topology, var_types=var_types, prob_categories=prob_cat, link_types=link_types,
                  snr=snr, treatment_importances=treatment_importance,
                  outcome_types=outcome_type, effect_sizes=self.no_X.effect_sizes,
                  treatment_methods=treatment_method,
                  survival_distribution=survival_distribution, survival_baseline=survival_baseline)
        num_samples = self.NUM_SAMPLES
        X, prop, cf = sim.generate_data(num_samples=num_samples, random_seed=783454)
        self.assertAlmostEqual(np.abs(X[4].le(X[3]).sum() / num_samples), prob_cat[3][0], places=1)
        # df_obs, df_cf = sim.format_for_training(X, prop, cf)

        # binary censor
        topology = np.zeros((5, 5), dtype=bool)
        topology[2, 0] = topology[4, 0] = topology[2, 1] = topology[3, 1] = topology[4, 2] = topology[4, 3] = True
        var_types = ["covariate", "covariate", "treatment", "censor", "outcome"]
        link_types = ["linear"] * 5
        prob_cat = [None, None, [0.2, 0.8], [0.85, 0.15], None]
        outcome_type = "continuous"
        snr = 0.95
        treatment_importance = None
        treatment_method = "logistic"
        sim = CS3(topology=topology, var_types=var_types, prob_categories=prob_cat, link_types=link_types,
                  snr=snr, treatment_importances=treatment_importance,
                  outcome_types=outcome_type, effect_sizes=self.no_X.effect_sizes,
                  treatment_methods=treatment_method)
        num_samples = self.NUM_SAMPLES
        X, prop, cf = sim.generate_data(num_samples=num_samples)
        self.assertAlmostEqual(X[3].astype(int).sum() / num_samples, prob_cat[3][1])
        df_obs, df_cf = sim.format_for_training(X, prop, cf)
        self.assertEqual(X[3].astype(int).sum(), df_obs["y_4"].isnull().sum())

        # independent categorical censor
        topology = np.zeros((5, 5), dtype=bool)
        topology[3, 0] = topology[4, 0] = topology[3, 2] = topology[4, 2] = topology[4, 3] = True
        var_types = ["covariate", "covariate", "treatment", "censor", "outcome"]
        link_types = ["linear"] * 5
        prob_cat = [None, None, [0.2, 0.8], [0.85, 0.15], None]
        outcome_type = "continuous"
        snr = 0.95
        treatment_importance = None
        treatment_method = "logistic"
        survival_distribution = "expon"
        survival_baseline = 0.8
        sim = CS3(topology=topology, var_types=var_types, prob_categories=prob_cat, link_types=link_types,
                  snr=snr, treatment_importances=treatment_importance,
                  outcome_types=outcome_type, effect_sizes=self.no_X.effect_sizes,
                  treatment_methods=treatment_method,
                  survival_distribution=survival_distribution, survival_baseline=survival_baseline)
        num_samples = self.NUM_SAMPLES
        X, prop, cf = sim.generate_data(num_samples=num_samples)
        self.assertAlmostEqual(X[3].astype(int).sum() / num_samples, prob_cat[3][1])
        # df_obs, df_cf = sim.format_for_training(X, prop, cf)

        # independent survival censor
        topology = np.zeros((5, 5), dtype=bool)
        topology[3, 0] = topology[4, 0] = topology[3, 2] = topology[4, 2] = topology[4, 3] = True
        var_types = ["covariate", "covariate", "treatment", "censor", "outcome"]
        link_types = ["linear"] * 5
        prob_cat = [None, None, [0.2, 0.8], [0.85, 0.15], None]
        outcome_type = "survival"
        snr = 0.95
        treatment_importance = None
        treatment_method = "logistic"
        survival_distribution = "expon"
        survival_baseline = 0.8
        sim = CS3(topology=topology, var_types=var_types, prob_categories=prob_cat, link_types=link_types,
                  snr=snr, treatment_importances=treatment_importance,
                  outcome_types=outcome_type, effect_sizes=self.no_X.effect_sizes,
                  treatment_methods=treatment_method,
                  survival_distribution=survival_distribution, survival_baseline=survival_baseline)
        num_samples = 10000
        X, prop, cf = sim.generate_data(num_samples=num_samples)
        # self.assertAlmostEqual(X[4].le(X[3]).sum() / float(num_samples), prob_cat[3][0], places=1)
        # df_obs, df_cf = sim.format_for_training(X, prop, cf)

        # TODO: test different link types
        # TODO: test marginal structural model (both in continuous, dichotomous and probability settings)


if __name__ == "__main__":
    unittest.main()



# (C) Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Created on Sept 25, 2019

"""
IHDP Data Downloading, Unzipping and Loading

This module provides data suitable for testing the HEMM estimator and for 
writing example notebooks.
"""

import numpy as np
import os
import urllib.request
import zipfile


def __download_data(url_path, local_path, verbose=0):
    if not os.path.exists(local_path):
        if verbose:
            print(f"Downloading data from {url_path} to {local_path}")
        req = urllib.request.urlretrieve(url=url_path, filename=local_path)
        return req[0]
    return local_path


def loadIHDPData(cache_dir=None, verbose=0, delete_extracted=True):
    """Downloads and loads IHDP-1000 dataset.
    Taken From Fredrik Johansson's website: http://www.fredjo.com/

    Args:
        cache_dir (str): Directory to which files will be downloaded
            If None: files will be downloaded to ~/causallib-data/.
        verbose (int): Controls the verbosity: the higher, the more messages.
        delete_extracted (bool): Delete extracted files from disk once loaded

    Returns:
        dict[str, dict[str, np.ndarray]]: "TRAIN" and "TEST" sets as keys.
            Values are dictionaries with `'x', 't', 'yf', 'ycf', 'mu0', 'mu1'` keys standing for
            covariates, treatment, factual outcome, counterfactual outcome, and noiseless potential outcomes

    Notes:
        Requires internet connection in case local data files do not already exist.
        Will save a local copy of the download

    """
    base_remote_url = "http://www.fredjo.com/files/"
    file_name = "ihdp_npci_1-1000.{phase}.npz.zip"

    # Set local download location:
    if cache_dir is None:
        cache_dir = os.path.join("~", 'causallib-data')
        cache_dir = os.path.expanduser(cache_dir)  # Expand ~ component to full path
        cache_dir = os.path.join(cache_dir, "IHDP")
    # cache_dir = cache_dir.replace("/", os.sep)
    os.makedirs(cache_dir, exist_ok=True)

    data = {}
    for phase in ["train", "test"]:
        # Obtain local copy of the data:
        phase_file_name = file_name.format(phase=phase)
        file_path = __download_data(
            url_path=base_remote_url + phase_file_name,
            local_path=os.path.join(cache_dir, phase_file_name),
            verbose=verbose
        )

        # Extract zipped data:
        npz_file_path = file_path.rsplit(".", maxsplit=1)[0]  # Remove ".zip" extension
        if not os.path.exists(npz_file_path):
            with zipfile.ZipFile(file_path) as zf:
                if verbose:
                    print(f"Extracting file into {npz_file_path}")
                zf.extractall(path=cache_dir)

        # Load data:
        phase_data = np.load(npz_file_path)
        phase_data = dict(phase_data)  # Load into memory, avoid lazy-loading
        data[phase.upper()] = phase_data

        # # In-memory extraction, works only in python>=3.7 https://github.com/python/cpython/pull/4966
        # with zipfile.ZipFile(file_path) as zf:
        #     internal_file_name = phase_file_name.rsplit(".", maxsplit=1)[0]  # Remove ".zip" extension
        #     with zf.open(internal_file_name, 'r') as npz_file:
        #         data[phase.upper()] = dict(np.load(npz_file))

        if delete_extracted:
            if verbose:
                print(f"Deleting extracted file {npz_file_path}")
            os.remove(npz_file_path)

    return data



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Jun 27, 2018

General (i.e. non-scientific) utils used throughout the package.
"""
from numpy import isscalar as np_is_scalar
from pandas import Series


def get_iterable_treatment_values(treatment_values, treatment_assignment, sort=True):
    """
    Convert an optionally provided specification of unique treatment values to an iterable of the unique treatment
    options.
    Since user can provide treatment values as either an iterable or a single value, this conversion to an iterable
    allows a generic approach of going over all provided treatment values.

    Args:
        treatment_values (None|Any|list[Any]): Unique values of possible treatment values.
                                               Can be either one value (scalar) or list of values (any iterable).
                                               Can be None, if None - treatment values are inferred from treatment
                                               assignment.
        treatment_assignment (Series): The observed treatment assignment, used to infer a list of unique treatment
                                       values in case no treatment values are provided (None is passed to
                                       treatment_values).
        sort (bool): Whether to sort the treatment values

    Returns:
        list[Any]: list of unique treatment values.
    """
    treatment_values = treatment_assignment.unique() if treatment_values is None else treatment_values
    treatment_values = [treatment_values] if np_is_scalar(treatment_values) else treatment_values
    if sort:
        treatment_values = sorted(treatment_values)
    return treatment_values


def create_repr_string(o):
    """

    Args:
        o (object): any core object

    Returns:
        str: repr string based on internal attributes
    """
    # Filter peripheral unimportant attribute names:
    params = [
        attr for attr in dir(o) if not attr.startswith('__')  # Data-model dunder methods
                                   and not callable(getattr(o, attr, None))  # Remove other methods, keep only fields
                                   and not attr.startswith("_abc")  # Remove abstract-related attributes
                                   and not attr.endswith("_")  # Remove attributes stated after initialization
                                   and not attr == "CALCULATE_EFFECT"  # Remove the EffectEstimator attribute
    ]

    # Special treatment for scikit-learn's learner object (the "learner" attribute) - place last in a new line:
    learner_included = False
    if "learner" in params:
        params.remove("learner")
        # params.append("learner")    # move learner to be last parameter
        learner_included = True

    # Couple attribute name with attribute value
    params = [(attr, getattr(o, attr)) for attr in params]
    params_string = ", ".join("{}={}".format(*param) for param in params)

    if learner_included:
        # Place "learner" attribute last in a different line:
        params_string += ",\n{spaces}learner={learner}".format(spaces=" " * (len(o.__class__.__name__) + 1),
                                                               learner=getattr(o, "learner"))
    repr_string = "{cls_name}({params})".format(cls_name=o.__class__.__name__,
                                                params=params_string)
    return repr_string


def check_learner_is_fitted(learner):
    """Return True if fitted and False otherwise"""
    # Following scikit-learn's convention,
    # fitted models have additional attributes ending with underscores.
    # See: https://scikit-learn.org/dev/glossary.html#term-fitted
    # Hence checking whether these exist is sufficient:
    after_init_attr = [attr for attr in learner.__dict__.keys() if attr.endswith("_")]
    is_fitted = len(after_init_attr) > 0
    return is_fitted



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Apr 25, 2018

"""

from typing import Mapping
import inspect

import pandas as pd
from numpy import isscalar
from sklearn.base import clone as sk_clone
from sklearn.preprocessing import OneHotEncoder

from .base_estimator import IndividualOutcomeEstimator
from ..utils import general_tools as g_tools


def _standardization_predict(estimator, X, predict_proba):
    """
    Single prediction call.

    Args:
        estimator: Fitted scikit estimator to use for prediction.
        X (pd.DataFrame): Data to predict on.
        predict_proba (bool): If applicable in the estimator (classification model) and if True - predict a continuous
                              value utilizing `predict_proba` or `decision_function`, rather than classifying with
                              `predict`.

    Returns:
        pd.DataFrame | pd.Series: If regression model or predict_proba=False then it returns a vector-like array (with
                                  prediction for each sample). If classification and predict_proba=True then returns
                                  a matrix-like array of n_samples by n_classes.
    """
    # Predict continuous values for classification if desired (predict_proba=True) and applicable (model has
    # `predict_proba` or `decision_function` methods to utilize:
    if predict_proba and hasattr(estimator, "predict_proba"):
        prediction = estimator.predict_proba(X)
    elif predict_proba and hasattr(estimator, "decision_function"):
        prediction = estimator.decision_function(X)
    else:
        prediction = estimator.predict(X)

    # Wrap results in pandas indexed array:
    if len(prediction.shape) == 1:  # A vector:
        prediction = pd.Series(prediction, index=X.index, name="y")
    else:  # A matrix:
        prediction = pd.DataFrame(prediction, index=X.index, columns=estimator.classes_)
        prediction.columns.names = ["y"]
    return prediction


def _add_sample_weight_fit_params(estimator, sample_weight):
    """Return fit params according to whether estimator is a simple estimator or a pipeline"""
    is_pipeline = hasattr(estimator, "steps")
    if is_pipeline:
        # Attribute the provided sample_weights to the final estimator in the pipeline.
        # Attribution is done by step name followed by dunder, see:
        # https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.Pipeline.html
        estimator_name, estimator = estimator.steps[-1]
        fit_params = {"{}__sample_weight".format(estimator_name): sample_weight}
    else:
        fit_params = dict(sample_weight=sample_weight)

    if "sample_weight" not in inspect.signature(estimator.fit).parameters and sample_weight is None:
        # Estimator does not support "sample_weight" parameter and sample_weight is not provided
        fit_params = {}

    return fit_params


class StratifiedStandardization(IndividualOutcomeEstimator):
    """
    Standardization model that learns a model for each treatment group (i.e. subgroup of subjects with the same
    treatment assignment).
    """

    def __init__(self, learner, treatment_values=None, predict_proba=False):
        """

        Args:
            learner: Initialized sklearn model or a mapping (dict) between treatment value and initialized model,
                      For example: {0: Ridge(alpha=5), 1: Ridge(alpha=0.1)},
                      or even different models all over: {0: Ridge(), 1: RandomForestRegressor}
                      Make sure these treatment_values keys represent all treatment values found in later use.
            treatment_values (list): list of unique values of treatment (can be a single value as well).
                                     If known beforehand (on initialization time), can be passed now to init, otherwise
                                     would be inferred during fit (where treatment assignment must be supplied).
                                     Make sure these treatment_values represent all treatment values found in later use.
            predict_proba (bool): In case the outcome task is classification and in case `learner` supports the
                                  operation, if True - prediction will utilize learner's `predict_proba` or
                                  `decision_function` which returns a continuous matrix of size (n_samples, n_classes).
                                  If False - `predict` will be used and return value will be based on a vector of class
                                  classifications.
        """
        super(StratifiedStandardization, self).__init__(learner, predict_proba=predict_proba)
        if isinstance(learner, Mapping):
            self.learner = learner
        elif treatment_values is not None:  # overwrite native `learner` with dictionary based one
            self.learner = self._clone_learner(treatment_values)
        self.treatment_values = treatment_values

    def _clone_learner(self, treatment_values):
        """
        Create a copy of underlining learner object for each of the treatment values.
        Args:
            treatment_values: lLst of unique values of treatment (can be a single value and not a list as well).

        Returns:
            dict[Any, learner]: Dictionary that holds for each treatment value (key) a learner object (value) that
                                was passed during initialization.
        """
        treatment_values = [treatment_values] if isscalar(treatment_values) else treatment_values
        learners = {treatment_value: sk_clone(self.learner) for treatment_value in treatment_values}
        return learners

    def _predict(self, X, treatment_value, predict_proba=None):
        """

        Args:
            X (pd.DataFrame): Data to predict on.
            treatment_value: What model to use. Each treatment value has its own trained model. `treatment_value` is
                             used to retrieve the appropriate model.
            predict_proba (bool): In case the outcome task is classification and in case `learner` supports the
                                  operation, if True - prediction will utilize learner's `predict_proba` or
                                  `decision_function` which returns a continuous matrix of size (n_samples, n_classes).
                                  If False - `predict` will be used and return value will be based on a vector of class
                                  classifications.

        Returns:
            pd.DataFrame | pd.Series: If regression model or predict_proba=False then it returns a vector-like array
                                      (with prediction for each sample). If classification and predict_proba=True then
                                      returns a matrix-like array of n_samples by n_classes.
        """
        predict_proba = self.predict_proba if predict_proba is None else predict_proba
        prediction = _standardization_predict(estimator=self.learner[treatment_value], X=X,
                                              predict_proba=predict_proba)
        return prediction

    def estimate_individual_outcome(self, X, a, treatment_values=None, predict_proba=None):
        treatment_values = g_tools.get_iterable_treatment_values(treatment_values, a)

        res = {}
        for treatment_value in treatment_values:
            prediction = self._predict(X=X, treatment_value=treatment_value, predict_proba=predict_proba)
            res[treatment_value] = prediction
        # TODO: should combine the results by the observed treatment into additional vector?
        res = pd.concat(res, axis="columns", names=[a.name or "a"])
        return res

    def fit(self, X, a, y, sample_weight=None):
        self.treatment_values_ = g_tools.get_iterable_treatment_values(None, a)
        if not isinstance(self.learner, dict):  # sk-learner was not cloned yet to have one copy for each stratum
            self.learner = self._clone_learner(self.treatment_values_)

        for cur_X, cur_y, cur_sw, treatment_value in self._prepare_data(X, a, y, sample_weight):
            fit_params = _add_sample_weight_fit_params(self.learner[treatment_value], cur_sw)
            self.learner[treatment_value] = self.learner[treatment_value].fit(cur_X, cur_y, **fit_params)
        return self

    def __repr__(self):
        # Since the learner is the same one only duplicated, it is redundant to repeat printing it.
        # Therefore, repeat it only once. And since there are no other attributes to consider, it is easy.
        repr_string = "{cls_name}(learner={params})".format(cls_name=self.__class__.__name__,
                                                            params=next(iter(self.learner.values())))
        return repr_string

    # def predict(self, X, a, treatment_values=None):
    #     res = self.estimate_individual_outcome(X, a, treatment_values)
    #     return res

    @staticmethod
    def _prepare_data(X, a, y=None, w=None):
        """
        Manipulating the data to fit the model specifications.
        This methods iterates of different treatment values, slices out the subgroups that were assigned to the
        specific treatment and yields the relevant dataset.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series | None): Observed outcome of size (num_subjects,).
            w (pd.Series | None): sample_weights

        Yields:
            (pd.DataFrame, pd.Series, Any): A three-tuple containing:

             * the covariates for individual under specific treatment,
             * the observed outcomes for these individuals (if y was passed and is not None),
             * the current treatment value.
        """
        treatment_values = g_tools.get_iterable_treatment_values(None, a)
        for treatment_value in treatment_values:
            treated = a == treatment_value
            cur_X = X.loc[treated, :]
            cur_y = y[treated] if y is not None else None
            cur_w = w[treated] if w is not None else None
            yield cur_X, cur_y, cur_w, treatment_value


class Standardization(IndividualOutcomeEstimator):
    """
    Standard standardization model for causal inference.
    Learns a model that takes into account the treatment assignment, and later, this value can be intervened, changing
    the predicted outcome.
    """

    def __init__(self, learner, encode_treatment=False, predict_proba=False):
        """

        Args:
            learner: Initialized sklearn model.
            encode_treatment (bool): Whether to encode the treatment as one-hot matrix.
                                     Usually good if n_treatment > 2.
            predict_proba (bool): In case the outcome task is classification and in case `learner` supports the
                                  operation, if True - prediction will utilize learner's `predict_proba` or
                                  `decision_function` which returns a continuous matrix of size (n_samples, n_classes).
                                  If False - `predict` will be used and return value will be based on a vector of class
                                  classifications.
        """
        super(Standardization, self).__init__(learner, predict_proba=predict_proba)
        self.encode_treatment = encode_treatment

    def estimate_individual_outcome(self, X, a, treatment_values=None, predict_proba=None):
        treatment_values = g_tools.get_iterable_treatment_values(treatment_values, a)
        res = {}
        for treatment_value in treatment_values:
            treatment_assignment = pd.Series(treatment_value, index=X.index)  # a vector of treatment assignment
            prediction = self._predict(X, treatment_assignment, predict_proba=predict_proba)  # predict
            res[treatment_value] = prediction  # Save prediction
        res = pd.concat(res, axis="columns", names=[a.name or "a"])
        return res

    # def _initialize_encoder(self, treatment_values):
    #     self.treatment_encoder_ = OneHotEncoder(n_values=len(treatment_values), sparse=False)

    def fit(self, X, a, y, sample_weight=None):
        if self.encode_treatment:
            # setattr(self, "treatment_encoder_", OneHotEncoder(sparse=False))
            self.treatment_encoder_ = OneHotEncoder(sparse=False, categories="auto")
            self.treatment_encoder_.fit(a.values.reshape(-1, 1))
        X = self._prepare_data(X, a)
        fit_params = _add_sample_weight_fit_params(self.learner, sample_weight)
        self.learner.fit(X, y, **fit_params)
        return self

    def _predict(self, X, a, predict_proba=None):
        """

        Args:
            X (pd.DataFrame): Covariates to predict on.
            a (pd.Series): Corresponding treatment assignment to utilize for prediction.
            predict_proba (bool): In case the outcome task is classification and in case `learner` supports the
                                  operation, if True - prediction will utilize learner's `predict_proba` or
                                  `decision_function` which returns a continuous matrix of size (n_samples, n_classes).
                                  If False - `predict` will be used and return value will be based on a vector of class
                                  classifications.

        Returns:
            pd.DataFrame | pd.Series: If regression model or predict_proba=False then it returns a vector-like array
                                      (with prediction for each sample). If classification and predict_proba=True then
                                      returns a matrix-like array of n_samples by n_classes.
        """
        predict_proba = self.predict_proba if predict_proba is None else predict_proba

        cur_X = self._prepare_data(X, a)  # concatenate treatment assignment to data

        prediction = _standardization_predict(estimator=self.learner, X=cur_X, predict_proba=predict_proba)
        return prediction

    def _prepare_data(self, X, a):
        """
        Manipulating the data to fit the model specifications.
        This method concatenates the treatment assignment (either a vector or a One-Hot matrix representing that vector)
        to the covariates and then learns a model on that augmented design matrix.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).

        Returns:
            pd.DataFrame: concatenation of treatment column/s to the provided covariate matrix (A | X).
        """
        if self.encode_treatment:
            a_transformed = self.treatment_encoder_.transform(a.values.reshape(-1, 1))
            a_name = a.name or "a"
            a = pd.DataFrame(a_transformed, index=a.index, columns=self.treatment_encoder_.categories_)
            a = a.add_prefix(a_name + "_")
        cur_X = pd.concat([a, X], join="outer", axis="columns")
        return cur_X



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Apr 25, 2018

"""

import pandas as pd

from .base_weight import WeightEstimator
from .base_estimator import PopulationOutcomeEstimator


class MarginalOutcomeEstimator(WeightEstimator, PopulationOutcomeEstimator):
    """
    A marginal outcome predictor.
    Assumes the sample is marginally exchangeable, and therefore does not correct (adjust, control) for covariates.
    Predicts the outcome/effect as if the sample came from a randomized control trial: $\\Pr[Y|A]$.
    """

    def compute_weight_matrix(self, X, a, use_stabilized=None, **kwargs):
        # Another way to view this is that Uncorrected is basically an IPW-like with all individuals equally weighted.
        treatment_values = a.unique()
        treatment_values = treatment_values.sort()
        weights = pd.DataFrame(data=1, index=a.index, columns=treatment_values)
        return weights

    def compute_weights(self, X, a, treatment_values=None, use_stabilized=None, **kwargs):
        # Another way to view this is that Uncorrected is basically an IPW-like with all individuals equally weighted.
        weights = pd.Series(data=1, index=a.index)
        return weights

    def fit(self, X=None, a=None, y=None):
        """
        Dummy implementation to match the API.
        MarginalOutcomeEstimator acts as a WeightEstimator that weights each sample as 1

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).

        Returns:
            MarginalOutcomeEstimator: a fitted model.
        """
        return self

    def estimate_population_outcome(self, X, a, y, w=None, treatment_values=None):
        """
        Calculates potential population outcome for each treatment value.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).
            w (pd.Series | None): Individual (sample) weights calculated. Used to achieved unbiased average outcome.
                                   If not provided, will be calculated on the data.
            treatment_values (Any): Desired treatment value/s to stratify upon before aggregating individual into
                                    population outcome.
                                    If not supplied, calculates for all available treatment values.

        Returns:
            pd.Series[Any, float]: Series which index are treatment values, and the values are numbers - the
                                   aggregated outcome for the strata of people whose assigned treatment is the key.
        """
        if w is None:
            w = self.compute_weights(X, a)
        res = self._compute_stratified_weighted_aggregate(y, sample_weight=w, stratify_by=a,
                                                          treatment_values=treatment_values)
        return res




"""
Methods for evaluating causal inference models.

(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Dec 25, 2018

"""


import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedKFold

from .predictions import WeightEvaluatorScores
from .predictor import predict_cv
from .results import EvaluationResults
from .scoring import score_cv

def _make_dummy_cv(n_samples):
    phases = ["train"]  # dummy phase
    # All DataFrame rows when using iloc
    cv = pd.RangeIndex(start=0, stop=n_samples)
    cv = [(cv, cv)]  # wrap in a tuple format compatible with sklearn's cv output
    return phases, cv


def _make_bootstrap_cv(n_samples_total, n_bootstrap, n_samples_bootstrap, replace):
    # Evaluation using bootstrap
    phases = ["train"]  # dummy phase

    # Generate bootstrap sample:
    cv = []
    # All DataFrame rows when using iloc
    X_ilocs = pd.RangeIndex(start=0, stop=n_samples_total)
    for _ in range(n_bootstrap):
        # Get iloc positions of a bootstrap sample (sample the size of X with replacement):
        # idx = X.sample(n=X.shape[0], replace=True).index
        # idx = np.random.random_integers(low=0, high=X.shape[0], size=X.shape[0])
        idx = np.random.choice(X_ilocs, size=n_samples_bootstrap, replace=replace)
        # wrap in a tuple format compatible with sklearn's cv output
        cv.append((idx, idx))
    return phases, cv


def evaluate(
    estimator,
    X,
    a,
    y,
    cv=None,
    metrics_to_evaluate="defaults",
    plots=False,
):
    """Evaluate model in cross-validation of the provided data

    Args:
        estimator (causallib.estimation.base_estimator.IndividualOutcomeEstimator |
            causallib.estimation.base_weight.WeightEstimator |
            causallib.estimation.base_weight.PropensityEstimator) : an estimator. If using cv, it
            will be refit, otherwise it should already be fit.
        X (pd.DataFrame): Covariates.
        a (pd.Series): Treatment assignment.
        y (pd.Series): Outcome.
        cv (list[tuples] | generator[tuples] | None): list the number of folds containing tuples of
            indices (train_idx, validation_idx) in an iloc manner (row number).
            If None, there will be no cross-validation. If `cv="auto"`, a stratified Kfold with
            5 folds will be created and used for cross-validation.
        metrics_to_evaluate (dict | "defaults" | None): key: metric's name, value: callable that
            receives true labels, prediction, and sample_weights (the latter may be ignored).
            If `"defaults"`, default metrics are selected. If `None`, no metrics are evaluated.
        plots (bool): whether to generate plots

    Returns:
        EvaluationResults
    """

    if cv is None:
        return _evaluate_simple(estimator, X, a, y, metrics_to_evaluate, plots)
    # when evaluate_cv gets cv=None it makes an auto cv so turn "auto" to None
    if cv == "auto":
        cv = None

    return _evaluate_cv(
        estimator=estimator,
        X=X,
        a=a,
        y=y,
        cv=cv,
        refit=True,
        phases=("train", "valid"),
        metrics_to_evaluate=metrics_to_evaluate,
        plots=plots,
    )


def _evaluate_simple(estimator, X, a, y, metrics_to_evaluate=None, plots=False):
    """Evaluate model on the provided data without cross-validation or bootstrap.

    Simple evaluation without cross validation on the provided data can be to test
    the model on its train data or on new data.

    Args:
        X (pd.DataFrame): Covariates.
        a (pd.Series): Treatment assignment.
        y (pd.Series): Outcome.
        metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives
            true labels, prediction and sample_weights (the latter is allowed to be ignored).
            If not provided, defaults from causallib.evaluation.metrics are used.
        plots (list[str] | None): list of plots to make. If None, none are generated.

    Returns:
        causallib.evaluation.results.EvaluationResults
    """

    phases, cv = _make_dummy_cv(X.shape[0])
    results = _evaluate_cv(
        estimator,
        X,
        a,
        y,
        cv=cv,
        refit=False,
        phases=phases,
        metrics_to_evaluate=metrics_to_evaluate,
        plots=plots,
    )

    results.remove_spurious_cv()

    return results


def _evaluate_cv(
    estimator,
    X,
    a,
    y,
    cv=None,
    refit=True,
    phases=("train", "valid"),
    metrics_to_evaluate=None,
    plots=False,
):
    """Evaluate model in cross-validation of the provided data

    Args:
        X (pd.DataFrame): Covariates.
        a (pd.Series): Treatment assignment.
        y (pd.Series): Outcome.
        cv (list[tuples] | generator[tuples]): list the number of folds containing tuples of
            indices (train_idx, validation_idx) in an iloc manner (row number).
        refit (bool): Whether to refit the model on each fold.
        phases (list[str]): {["train", "valid"], ["train"], ["valid"]}.
            Phases names to evaluate on - train ("train"), validation ("valid") or both.
            'train' corresponds to cv[i][0] and 'valid' to  cv[i][1]
        metrics_to_evaluate (dict | None): key: metric's name, value: callable that
            receives true labels, prediction, and sample_weights (the latter may be ignored).
        plots (bool): whether to generate plots

    Returns:
        EvaluationResults
    """
    # We need consistent splits for predicting, scoring and plotting.
    # If cv is a generator, it would be lost after after first use.
    # If kfold has shuffle=True, it would be inconsistent.
    # To keep consistent reproducible folds, we save them as a list.
    if cv is None:
        kfold = StratifiedKFold(n_splits=5)
        cv = kfold.split(X=X, y=a)

    # if cv is generator it would listify it, if cv is already a list this is idempotent
    cv = list(cv)

    predictions, models = predict_cv(estimator, X, a, y, cv, refit, phases)
    if metrics_to_evaluate is not None:
        evaluated_metrics = score_cv(predictions, X, a, y, cv, metrics_to_evaluate)
    else:
        evaluated_metrics = None
    evaluation_results = EvaluationResults.make(
        evaluated_metrics=evaluated_metrics,
        predictions=predictions,
        cv=cv,
        models=models if refit is True else [estimator],
        X=X,
        a=a,
        y=y,
    )

    if plots:
        evaluation_results.plot_all()
    return evaluation_results


def evaluate_bootstrap(
    estimator,
    X,
    a,
    y,
    n_bootstrap,
    n_samples=None,
    replace=True,
    refit=False,
    metrics_to_evaluate=None,
):
    """Evaluate model on a bootstrap sample of the provided data

    Args:
        X (pd.DataFrame): Covariates.
        a (pd.Series): Treatment assignment.
        y (pd.Series): Outcome.
        n_bootstrap (int): Number of bootstrap sample to create.
        n_samples (int | None): Number of samples to sample in each bootstrap sampling.
            If None - will use the number samples (first dimension) of the data.
        replace (bool): Whether to use sampling with replacements.
            If False - n_samples (if provided) should be smaller than X.shape[0])
        refit (bool): Whether to refit the estimator on each bootstrap sample.
            Can be computational intensive if n_bootstrap is large.
        metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives
            true labels, prediction and sample_weights (the latter is allowed to be ignored).
            If not provided, default from causallib.evaluation.metrics are used.

    Returns:
        EvaluationResults
    """
    n_samples_total = X.shape[0]

    if n_samples is None:
        n_samples = n_samples_total

    phases, cv = _make_bootstrap_cv(n_samples_total, n_bootstrap, n_samples, replace)

    results = _evaluate_cv(
        estimator,
        X,
        a,
        y,
        cv=cv,
        refit=refit,
        phases=phases,
        metrics_to_evaluate=metrics_to_evaluate,
        plots=None,
    )

    
    results.remove_spurious_cv()
    if results.evaluated_metrics is not None:
        if isinstance(results.evaluated_metrics, WeightEvaluatorScores):
            results.evaluated_metrics.covariate_balance.index.rename("sample", "fold", inplace=True)
            results.evaluated_metrics.prediction_scores.index.rename("sample", "fold", inplace=True)
        else:
            results.evaluated_metrics.index.rename("sample", "fold", inplace=True)
    
    return results



"""
Metrics to assess performance of propensity score models.
Function named as ``*_error`` return a scalar value to minimize:
the lower, the better.

Metrics' interface doesn't strictly follow skleran's metrics interface.
Names are kept as `y_true` and `y_pred` even though these regard the
treatment assignment `a` (`a_true`, `a_pred` [propensity]).
"""

import numpy as np
from numpy import interp
from sklearn.metrics import roc_auc_score, roc_curve
import statsmodels.api as sm


def weighted_roc_auc_error(y_true, y_pred, sample_weight):
    """
    Compute the squared error between the balanced (e.g. IP-weighted) ROC AUC
    to the diagonal, i.e. AUC=0.5.

    Shimoni, Y., et al. (2019)
    An evaluation toolkit to guide model selection and cohort definition in causal inference.

    Args:
        y_true (pd.Series): True binary label assignment of size (num_subjects,)
        y_pred (pd.Series): Predicted probability of each sample being
                            the positive label of size (num_subjects,).
        sample_weight (pd.Series | None): Weights of size (num_subjects,)

    Returns:
        score (float):
    """
    weighted_auc = roc_auc_score(y_true, y_pred, sample_weight=sample_weight)
    chance_auc = 0.5
    score = (weighted_auc - chance_auc) ** 2
    return score


def expected_roc_auc_error(y_true, y_pred, **kwargs):
    """
    Compute the squared error between the expected ROC-AUC given the provided
    scores and the actual ROC-AUC they produce.

    Shimoni, Y., et al. (2019)
    An evaluation toolkit to guide model selection and cohort definition in causal inference.

    Args:
        y_true (pd.Series): True binary label assignment of size (num_subjects,)
        y_pred (pd.Series): Predicted probability of each sample being
                            the positive label of size (num_subjects,).

    Returns:
        score (float):
    """
    # Calculate expected roc auc:
    p = np.hstack((y_pred, y_pred))
    w = np.hstack((y_pred, 1 - y_pred))
    target = np.hstack((np.ones_like(y_pred), np.zeros_like(y_pred)))
    expected_auc = roc_auc_score(target, p, sample_weight=w)

    auc = roc_auc_score(y_true, y_pred)
    score = (expected_auc - auc) ** 2
    return score


def weighted_roc_curve_error(y_true, y_pred, sample_weight, agg=np.max):
    """Compute the absolute differences between the balanced (e.g. IP-weighted) ROC curve
    and the diagonal x=y curve.
    Since difference in curves results in a multiple values (each point along the curve),
    an aggregation function is also required.

    Shimoni, Y., et al. (2019)
    An evaluation toolkit to guide model selection and cohort definition in causal inference.

    Args:
        y_true (pd.Series): True binary label assignment of size (num_subjects,)
        y_pred (pd.Series): Predicted probability of each sample being
                            the positive label of size (num_subjects,).
        sample_weight (pd.Series | None): Weights of size (num_subjects,)
        agg (callable): A function to aggregate a vector of absolute differences
                        between the curves' points. Default is max.

    Returns:
        score (float):
    """
    weighted_roc_fpr, weighted_roc_tpr, _ = roc_curve(
        y_true, y_pred, sample_weight=sample_weight
    )
    diagonal = weighted_roc_fpr
    diff = np.abs(weighted_roc_tpr - diagonal)
    score = agg(diff)
    return score


def expected_roc_curve_error(y_true, y_pred, agg=np.max, **kwargs):
    """Compute the absolute differences between the expected ROC curve and the regular
    ROC curve of the model.
    Since difference in curves results in a multiple values (each point along the curve),
    an aggregation function is also required.

    Shimoni, Y., et al. (2019)
    An evaluation toolkit to guide model selection and cohort definition in causal inference.

    Args:
        y_true (pd.Series): True binary label assignment of size (num_subjects,)
        y_pred (pd.Series): Predicted probability of each sample being
                            the positive label of size (num_subjects,).
        agg (callable): A function to aggregate a vector of absolute differences
                        between the curves' points. Default is max.

    Returns:
        score (float):
    """
    # Calculate expected roc curve:
    p = np.hstack((y_pred, y_pred))
    w = np.hstack((y_pred, 1 - y_pred))
    target = np.hstack((np.ones_like(y_pred), np.zeros_like(y_pred)))
    expected_fpr, expected_tpr, _ = roc_curve(target, p, sample_weight=w)

    fpr, tpr, _ = roc_curve(y_true, y_pred)

    # Align the FPR/TPR so the difference vector is meaningful:
    fpr_domain = np.linspace(0, 1, 100)
    tpr_interp = interp(fpr_domain, fpr, tpr)
    tpr_interp[0] = 0.0
    expected_tpr_interp = interp(fpr_domain, expected_fpr, expected_tpr)
    expected_tpr_interp[0] = 0.0

    diff = np.abs(tpr_interp - expected_tpr_interp)
    score = agg(diff)
    return score


def ici_error(y_true, y_pred, agg=np.mean, lowess_kwargs=None, **kwargs):
    """Integrated calibration index metric.
    Fits a lowess model between binary treatment assignment and the predicted probabilities.
    This generates a smooth calibration curve,
    which we can subtract from the diagonal, and examine magnitude.
    To make a single-number score the absolute difference is then aggregated using mean, median, max,
    or whatever collapses a vector into a single number.

    See: Austin and Steyerberg (2019): The Integrated Calibration Index (ICI)
    and related metrics for quantifying the calibration of logistic regression models.
    """
    if lowess_kwargs is None:
        lowess_kwargs = {}

    smooth_curve = sm.nonparametric.lowess(
        y_true, y_pred,
        **lowess_kwargs
        # TODO: maybe force prediction domain xval=linspace(0, 1, num=0) or linspace(y_true.min(), y_true.max(), num=0)
    )
    # See `out` in doc: https://www.statsmodels.org/dev/generated/statsmodels.nonparametric.smoothers_lowess.lowess.html
    if len(smooth_curve.shape) == 2:  # `return_sorted==True`
        y_pred = smooth_curve[:, 0]
        smooth_curve = smooth_curve[:, 1]

    diff = np.abs(smooth_curve - y_pred)
    score = agg(diff)
    return score





"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
import abc

import numpy as np
import pandas as pd
import scipy.stats as stats
import sklearn.feature_selection as feature_selection
from sklearn.base import BaseEstimator, TransformerMixin

from ..utils.stat_utils import isBinary, areColumnsBinary, computeCorrPvals


def track_selected_features(pipeline_stages, num_features):
    """

    Args:
        pipeline_stages (list [tuple[str, TransformerMixin]]): list of steps. each step is a tuple of Name and
                                                               Transformer Object.
        num_features (int):

    Returns:
        np.ndarray:
    """
    selected_features = np.arange(num_features)
    for p_name, p in pipeline_stages:
        if not isinstance(p, BaseFeatureSelector):
            continue
        p_features = p.selected_features
        selected_features = selected_features[p_features]
    return selected_features


class BaseFeatureSelector(BaseEstimator, TransformerMixin):
    """

    """

    def __init__(self):
        """

        """
        self._selected_features = None

    @property
    def selected_features(self):
        return self._selected_features

    @selected_features.setter
    def selected_features(self, features):
        if np.sum(self.selected_features) == 0:
            raise AssertionError("All features were removed by feature")
        self._selected_features = features

    def transform(self, X):
        """

        Args:
            X (pd.DataFrame):

        Returns:
            pd.DataFrame:
        """
        return X.loc[:, self.selected_features]

    @abc.abstractmethod
    def fit(self, X, y=None):
        """

        Args:
            X (pd.DataFrame): array-like, shape [n_samples, n_features] The data used for filtering.
            y: Passthrough for ``Pipeline`` compatibility.

        Returns:
            BaseFeatureSelector
        """
        raise NotImplementedError


class ConstantFilter(BaseFeatureSelector):
    """Removes features that are almost constant"""

    def __init__(self, threshold=0.95):
        """

        Args:
            threshold (float):
        """
        super(ConstantFilter, self).__init__()
        self.threshold = threshold

    def fit(self, X, y=None):
        is_const = pd.Series(0, index=X.columns, dtype=np.dtype(bool))
        for col in X.columns:
            # NaNs are not counted using unique (since np.nan != np.nan). Fill them with a unique value:
            cur_col = X.loc[:, col]
            cur_col.loc[~np.isfinite(cur_col)] = cur_col.max() + 1
            # Get values' frequency:
            freqs = cur_col.value_counts(normalize=True)
            is_const[col] = np.any(freqs > self.threshold)

        self.selected_features = ~is_const
        return self


class SparseFilter(BaseFeatureSelector):
    """Removes features with many missing values"""

    def __init__(self, threshold=0.2):
        """

        Args:
            threshold (float):
        """
        super(SparseFilter, self).__init__()
        self.threshold = threshold

    def fit(self, X, y=None):
        nan_freqs = np.mean(np.isnan(X), axis=0)
        is_sparse = nan_freqs > self.threshold
        self.selected_features = ~is_sparse
        return self


class HrlVarFilter(BaseFeatureSelector):
    """Removes features with a small variance, while allowing for missing values"""

    def __init__(self, threshold=0.0):
        """

        Args:
            threshold (float):
        """
        super(HrlVarFilter, self).__init__()
        self.threshold = threshold

    def fit(self, X, y=None):
        vars = np.nanvar(X, axis=0)
        self.selected_features = vars > self.threshold
        return self


class UnivariateAssociationFilter(BaseFeatureSelector):
    """Removes features according to univariate association"""

    def __init__(self, is_linear=True, threshold=0.2):
        """

        Args:
            is_linear (bool):
            threshold (float):
        """
        super(BaseFeatureSelector, self).__init__()
        self.is_linear = is_linear
        self.threshold = threshold

    def fit(self, X, y=None):
        p_vals = self.compute_pvals(X, y)

        self.selected_features = p_vals < self.threshold
        return self

    def compute_pvals(self, X, y):
        # TODO: export to stats_utils?
        is_y_binary = (len(np.unique(y)) == 2)
        # is_binary_feature = np.sum(((X != np.nanmin(X, axis=0)[np.newaxis, :]) &
        #                             (X != np.nanmax(X, axis=0)[np.newaxis, :])), axis=0) == 0
        is_binary_feature = areColumnsBinary(X)
        p_vals = np.zeros(X.shape[1])
        if is_y_binary:
            # Process non-binary columns:
            for i in np.where(~is_binary_feature)[0]:
                x0 = X.loc[y == 0, i]
                x1 = X.loc[y == 1, i]
                if self.is_linear:
                    _, p_vals[i] = stats.ttest_ind(x0, x1)
                else:
                    _, p_vals[i] = stats.ks_2samp(x0, x1)

            # Process binary features:
            _, p_vals[is_binary_feature] = feature_selection.chi2(X.loc[:, is_binary_feature], y)

        else:
            # Process non-binary features:
            _, p_vals[~is_binary_feature] = feature_selection.f_regression(X.loc[:, ~is_binary_feature], y)

            # Process binary features:
            y_mat = np.row_stack(y)
            for i in np.where(is_binary_feature)[0]:
                _, p_vals[i] = feature_selection.f_regression(y_mat, X.loc[:, i])
        return p_vals


class StatisticalFilter(BaseFeatureSelector):
    """Removes features according to univariate association"""

    # TODO: isn't this the same as the above?

    def __init__(self, threshold=0.2, isLinear=True):
        """

        Args:
            isLinear (bool):
            threshold (float):
        """
        super(StatisticalFilter, self).__init__()
        self.isLinear = isLinear
        self.threshold = threshold

    def fit(self, X, y=None):
        is_y_binary = isBinary(y)
        is_binary_feature = areColumnsBinary(X)
        p_vals = computeCorrPvals(X, y, is_binary_feature, is_y_binary, self.isLinear)

        self.selected_features = p_vals < self.threshold
        return self


class CorrelationFilter(BaseFeatureSelector):
    """Removes features that are strongly correlated to other features"""

    def __init__(self, threshold=0.9):
        """

        Args:
            threshold (float):
        """
        super(CorrelationFilter, self).__init__()
        self.threshold = threshold

    def fit(self, X, y=None):
        if np.any(np.isnan(X)):
            raise AssertionError("Correlation filter cannot handle NaN values")

        stat_filter = UnivariateAssociationFilter(threshold=1.0)  # Do not remove features prior the following process
        stat_filter.fit(X, y)
        p_vals = stat_filter.compute_pvals(X, y)
        # p_vals = computeCorrPvals(X, y)

        features_sorted = np.argsort(p_vals)
        is_removed = np.zeros(X.shape[1], dtype=np.bool)
        corr_mat = np.corrcoef(X.T)
        for i in features_sorted:  # iterate by p-values to keep the most significant among the highly correlated.
            if is_removed[i]:
                continue
            is_above_threshold = np.abs(corr_mat[:, i]) > self.threshold
            is_above_threshold[i] = False  # Ignore the correlation of i with itself.
            is_removed[is_above_threshold] = True  # Mark all features with big correlation to removal.

        self.selected_features = ~is_removed
        return self



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
from builtins import range

import numpy as np
import scipy.stats as stats
import sklearn.feature_selection as feature_selection
from pandas import DataFrame as pdDataFrame, Series as pdSeries  # For type hinting purposes only
from pandas.core.indexes.base import InvalidIndexError


def is_vector_binary(vec):
    return np.unique(vec).size <= 2


def which_columns_are_binary(X):
    """

    Args:
        X (pdDataFrame):

    Returns:

    """
    return X.apply(lambda x: x.nunique()).le(2)


def isBinary(x):
    """
    Asses whether a vector is binary.
    Args:
        x (pdSeries | np.ndarray):

    Returns:
        bool: True iff x is binary.
    """
    if hasattr(x, "values"):
        x = x.values
    x = x.reshape((-1, 1))
    return areColumnsBinary(x)[0]


def areColumnsBinary(X):
    """
    Assess whether all matrix columns are binary.
    Args:
        X (np.ndarray | pdDataFrame): Covariate matrix.

    Returns:
        np.ndarray: A boolean vector the length of number of features (columns). An entry is True iff the corresponding 
                    column is binary.
    """
    mins = np.nanmin(X, axis=0)  # min value for every column
    maxs = np.nanmax(X, axis=0)  # max value for every column
    X0 = X == mins[np.newaxis, :]  # TODO: newaxis is redundant?
    X1 = X == maxs[np.newaxis, :]  # TODO: newaxis is redundant?
    is_binary_features = np.all(X0 | X1, axis=0)  # binary vector stating which column is binary

    return is_binary_features


def computeCorrPvals(X, y, is_X_binary, is_y_binary, isLinear=True):
    """

    Args:
        X (pd.DataFrame): The covariate matrix
        y (pdSeries): The response
        is_X_binary (np.ndarray): Indication which columns are binary
        is_y_binary (bool): Indication whether the response vector is binary or not.
        isLinear (bool): Whether to perform a linear (slope) test (t-test) on the non-binary features or to perform
                         a two-sample Kolmogorov-Smirnov test

    Returns:
        np.array: A vector of p-values, one for every feature.
    """
    p_vals = np.empty(X.shape[1]) * np.nan
    y = np.squeeze(y)  # TODO: is it necessary?
    if is_y_binary:
        if any(
                is_X_binary):  # TODO: there's a better way, by first assigning Z=X[:, is_X_binary] and cheking whether Z is empty (or apply everything on an empty matrix)
            X2 = X.loc[:, is_X_binary]
            # xx = X2[:,[7]]
            #             t0 = timeit.default_timer()
            #             _, p1 = feature_selection.chi2(X2,y)
            #             t1 = timeit.default_timer()
            #             print 'after chi2_test_fs', (t1-t0)

            # t0 = timeit.default_timer()
            p2 = chi2_test(X2, y)
            # t1 = timeit.default_timer()
            # print 'after chi2_test', (t1-t0)

            p_vals[is_X_binary] = p2

        if (any(~is_X_binary)):
            # _, p_vals[is_X_binary] = feature_selection.f_classif(X[:, is_X_binary], y)
            for i in np.where(~is_X_binary)[0]:
                x = X.iloc[:, i]
                x0 = x[(~np.isnan(x)) & (y == 0)]  # TODO: y doesn;t have to be 0s or 1s
                x1 = x[(~np.isnan(x)) & (y == 1)]  # TODO: y doesn;t have to be 0s or 1s
                if isLinear:
                    _, p_vals[i] = stats.ttest_ind(x0, x1)
                else:
                    _, p_vals[i] = stats.ks_2samp(x0, x1)
    else:
        # if (any(~is_X_binary)):
        #    _, p_vals[~is_X_binary] = feature_selection.f_regression(X[:, ~is_X_binary], y)
        # y_mat = np.row_stack(y)
        # for i in np.where(is_X_binary)[0]:
        #    _, p_vals[i] = feature_selection.f_regression(y_mat, X[:, i])
        _, p_vals = feature_selection.f_regression(X, y)

    return p_vals


# def mycrosstab(x,y):
#     catx = np.unique(x)
#     caty = np.unique(y)
#     tab = [[sum((x==cx) & (y==cy)) for cx in catx] for cy in caty]
#     
#     #res1 = stats.chi2_contingency(tab, False)
#     res2 = stats.chi2_contingency(tab, True)
#     #return res1[1], res2[1]
#     return res2[1]

# def chi2_test_fs(X,y):
#     
#     m = X.shape[1]
#     pvals = np.empty(m)*np.NaN
#     for i in range(m):
#         Xi = X[:,[i]]
#         Xi01 = np.hstack((1-Xi, Xi))
#         _, pvals[i]= feature_selection.chi2(Xi01, y)
#     
#     return pvals


def chi2_test(X, y):
    """

    Args:
        X (np.ndarray): Binary feature matrix
        y (np.ndarray): Binary response vector

    Returns:
        np.ndarray: A vector of p-values, one for every feature.
    """
    X0 = 1 - X
    if hasattr(y, "values"):
        y = y.values
    Y = y.reshape((-1, 1))
    Y = np.append(1 - Y, Y, axis=1)
    Tbl1 = np.dot(Y.T, X)
    Tbl0 = np.dot(Y.T, X0)

    m = X.shape[1]
    pvals = np.empty(m) * np.NaN
    for i in range(m):
        if np.all([Tbl1[:, i] == 0]) or np.all([Tbl0[:, i] == 0]):
            pvals[i] = 1
        else:
            r = stats.chi2_contingency([Tbl0[:, i], Tbl1[:, i]], True)
            pvals[i] = r[1]
    return pvals


def calc_weighted_standardized_mean_differences(x, y, wx, wy, weighted_var=False):
    r"""
    Standardized mean difference: frac{\mu_1 - \mu_2 }{\sqrt{\sigma_1^2 + \sigma_2^2}}

    References:
        [1]https://cran.r-project.org/web/packages/cobalt/vignettes/cobalt_A0_basic_use.html#details-on-calculations
        [2]https://en.wikipedia.org/wiki/Strictly_standardized_mean_difference#Concept

    Note on variance:
    - The variance is calculated on unadjusted to avoid paradoxical situation when adjustment decreases both the
      mean difference and the spread of the sample, yielding a larger smd than that prior to adjustment,
      even though the adjusted groups are now more similar [1].
    - The denominator is as depicted in the "statistical estimation" section:
      https://en.wikipedia.org/wiki/Strictly_standardized_mean_difference#Statistical_estimation,
      namely, disregarding the covariance term [2], and is unweighted as suggested above in [1].
    """
    numerator = np.average(x, weights=wx) - np.average(y, weights=wy)
    if weighted_var:
        var = lambda vec, weights: np.average((vec - np.average(vec, weights=weights)) ** 2, weights=weights)
        denominator = np.sqrt(var(x, wx) + var(y, wy))
    else:
        denominator = np.sqrt(np.nanvar(x) + np.nanvar(y))
    if np.isfinite(denominator) and np.isfinite(numerator) and denominator != 0:
        bias = numerator / denominator
    else:
        bias = np.nan
    return bias


def calc_weighted_ks2samp(x, y, wx, wy):
    """
    Weighted Kolmogorov-Smirnov

    References:
        [1] https://stackoverflow.com/a/40059727
    """
    x_ix = np.argsort(x)
    y_ix = np.argsort(y)
    x, wx = x[x_ix], wx[x_ix]
    y, wy = y[y_ix], wy[y_ix]
    data = np.concatenate((x, y))
    wx_cum = np.hstack([0, wx.cumsum() / wx.sum()])
    wy_cum = np.hstack([0, wy.cumsum() / wy.sum()])
    # Align the "steps" between the two distribution so the differences will be well defined:
    x_align = wx_cum[[np.searchsorted(x, data, side="right")]]
    y_align = wy_cum[[np.searchsorted(y, data, side="right")]]
    stat = np.max(np.abs(x_align - y_align))
    # stat = ks_2samp(wx * x, wy * y)
    return stat


def robust_lookup(df, indexer):
    """
    Robust way to apply pandas lookup when indices are not unique

    Args:
        df (pdDataFrame):
        indexer (pdSeries): A Series whose index is either same or a subset of `df.index`
                            and whose values are values from `df.columns`.
                            If `a.index` contains values not in `df.index` 
                            they will have NaN values.

    Returns:
        pdSeries: a vector where (logically) `extracted[i] = df.loc[indexer.index[i], indexer[i]]`. 
            In most cases, when `indexer.index == df.index` this translates to 
            `extracted[i] = df.loc[i, indexer[i]]`
    """
    # Convert the index into 
    idx, col = indexer.factorize()  # convert text labels into integers
    extracted = df.reindex(col, axis=1).reindex(indexer.index, axis=0)  # make sure the columns exist and the indeces are the same
    extracted = extracted.to_numpy()[range(len(idx)), idx]  # numpy accesses by location, not by named index
    extracted = pdSeries(extracted, index=indexer.index)
    return extracted



"""
(C) IBM Corp, 2019, All rights reserved
Created on Aug 25, 2019

@author: EHUD KARAVANI
"""

import unittest
import pandas as pd
from causallib.estimation.base_weight import WeightEstimator


class TestBaseWeight(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.y = pd.Series([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        cls.w = pd.Series([1, 1, 1, 1, 1, 1, 1, 1, 0.5, 0.5])
        cls.a = pd.Series([0, 0, 0, 0, 1, 0, 1, 1, 1, 1])

    def setUp(self):
        self.model = WeightEstimator(learner=None)

    def test_no_weighting_no_stratification(self):
        result = self.model._compute_stratified_weighted_aggregate(self.y, None, None)
        truth = pd.Series(5/10, index=[0])
        pd.testing.assert_series_equal(truth, result)

    def test_weighting_no_stratification(self):
        result = self.model._compute_stratified_weighted_aggregate(self.y, self.w, None)
        truth = pd.Series(4/9, index=[0])
        pd.testing.assert_series_equal(truth, result)

    def test_no_weighting_stratification(self):
        result = self.model._compute_stratified_weighted_aggregate(self.y, None, self.a)
        truth = pd.Series([1/5, 4/5], index=[0, 1])
        pd.testing.assert_series_equal(truth, result)

    def test_weighting_stratification(self):
        result = self.model._compute_stratified_weighted_aggregate(self.y, self.w, self.a)
        truth = pd.Series([1/5, 3/4], index=[0, 1])
        pd.testing.assert_series_equal(truth, result)

    def test_subset_treatment_values(self):
        with self.subTest("Subset of treatment values exist in treatment"):
            result = self.model._compute_stratified_weighted_aggregate(self.y, None, self.a, [0])
            truth = pd.Series([1/5], index=[0])
            pd.testing.assert_series_equal(truth, result)

        with self.subTest("Subset of treatment values does not exist in treatment"):
            with self.assertRaises(ZeroDivisionError):  # Since the group is empty its weights' sum is zero
                self.model._compute_stratified_weighted_aggregate(self.y, None, self.a, [3])


